"""
Vector Store on Weavite.
"""

from __future__ import annotations

__all__ = ["Weaviate"]

import copy
from typing import Any

import weaviate
import weaviate.classes as wvc
from weaviate.classes.query import Filter

from verse.core import Context, NCall, Operation, Response
from verse.core.exceptions import BadRequestError, NotFoundError
from verse.internal.storage_core import (
    ItemProcessor,
    ParameterParser,
    QueryProcessor,
    StoreFunctionName,
    StoreOperation,
    StoreOperationParser,
    StoreProvider,
    Validator,
)
from verse.ql import (
    And,
    Comparison,
    ComparisonOp,
    Expression,
    Field,
    Function,
    FunctionNamespace,
    Not,
    Or,
    OrderBy,
    Select,
)

from .._helper import get_collection_config, is_metadata_field, is_value_field
from .._models import (
    VectorCollectionConfig,
    VectorIndexMetric,
    VectorIndexStructure,
    VectorItem,
    VectorKey,
    VectorList,
    VectorProperties,
    VectorSearchArgs,
    VectorValue,
)

DEFAULT_LIMIT = 100


class Weaviate(StoreProvider):
    type: str
    collection: str | None
    dimension: int | dict[str, int]
    content_properties_field: str | None
    url: str | None
    api_key: str | None
    http_host: str
    http_port: int
    http_secure: bool
    grpc_host: str
    grpc_port: int
    grpc_secure: bool
    nparams: dict[str, Any]

    _database_client: weaviate.WeaviateClient
    _init: bool

    _processor: ItemProcessor

    _collection_cache: dict[str, WeaviateCollection]

    def __init__(
        self,
        type: str = "wcs",
        collection: str | None = None,
        dimension: int | dict[str, int] = 4,
        content_properties_field: str | None = "_content",
        url: str | None = None,
        api_key: str | None = None,
        http_host: str = "localhost",
        http_port: int = 8080,
        http_secure: bool = False,
        grpc_host: str = "localhost",
        grpc_port: int = 50051,
        grpc_secure: bool = False,
        nparams: dict[str, Any] = dict(),
        **kwargs,
    ):
        """Initialize.

        Args:
            type:
                Weaviate connection type.
                Values are "wcs", "local", "embedded", "custom".
                Defaults to "wcs".
            collection:
                Collection name mapped to vector store collection.
            dimension:
                Vector dimension.
                To specify for multiple collections, use a dictionary
                where the key is the collection name and the value
                is the dimension.
            content_properties_field:
                Properties field where content is stored.
                Weaviate doesn't support a separate content field.
                Defaults to "_content".
            url:
                Weaviate url.
            api_key:
                Weaviate api key.
            http_host:
                Http host.
            http_port:
                Http port.
            http_secure:
                Secure http option.
            grpc_host:
                Grpc host.
            grpc_port:
                Grpc port.
            grpc_secure:
                Secure grpc option.
            nparams:
                Native parameters for weaviate client.
        """
        self.type = type
        self.collection = collection
        self.dimension = dimension
        self.content_properties_field = content_properties_field
        self.url = url
        self.api_key = api_key
        self.http_host = http_host
        self.http_port = http_port
        self.http_secure = http_secure
        self.grpc_host = grpc_host
        self.grpc_port = grpc_port
        self.grpc_secure = grpc_secure
        self.nparams = nparams

        self._init = False
        self._processor = ItemProcessor()
        self._collection_cache = dict()

    def init(self, context: Context | None = None) -> None:
        if self._init:
            return

        if (
            self.type == "wcs"
            and self.url is not None
            and self.api_key is not None
        ):
            self._database_client = weaviate.connect_to_wcs(
                cluster_url=self.url,
                auth_credentials=weaviate.auth.AuthApiKey(self.api_key),
                **self.nparams,
            )
        elif self.type == "local":
            self._database_client = weaviate.connect_to_local(
                **self.nparams,
            )
        elif self.type == "embedded":
            self._database_client = weaviate.connect_to_embedded(
                **self.nparams,
            )
        elif self.type == "custom":
            self._database_client = weaviate.connect_to_custom(
                http_host=self.http_host,
                http_port=self.http_port,
                http_secure=self.http_secure,
                grpc_host=self.grpc_host,
                grpc_port=self.grpc_port,
                grpc_secure=self.grpc_secure,
                **self.nparams,
            )
        else:
            raise BadRequestError("Cannot initalize client")
        self._init = True

    def _get_collection_name(
        self, op_parser: StoreOperationParser
    ) -> str | None:
        collection_name = (
            op_parser.get_operation_parsers()[0].get_collection_name()
            if op_parser.op_equals(StoreOperation.BATCH)
            or op_parser.op_equals(StoreOperation.TRANSACT)
            else op_parser.get_collection_name()
        )
        collection = (
            collection_name if collection_name is not None else self.collection
        )
        return collection

    def _get_collection(
        self, op_parser: StoreOperationParser
    ) -> WeaviateCollection | None:
        if op_parser.is_resource_op():
            return None
        collection_name = self._get_collection_name(op_parser)
        if collection_name is None:
            raise BadRequestError("Collection name must be specified")
        if collection_name in self._collection_cache:
            return self._collection_cache[collection_name]
        client = self._database_client.collections.get(collection_name)
        dimension = ParameterParser.get_collection_parameter(
            self.dimension, collection_name
        )
        col = WeaviateCollection(
            client,
            collection_name,
            dimension,
            self.content_properties_field,
        )
        self._collection_cache[collection_name] = col
        return col

    def _validate(self, op_parser: StoreOperationParser):
        if op_parser.op_equals(StoreOperation.BATCH):
            Validator.validate_batch(
                op_parser.get_operation_parsers(),
                allowed_ops=[
                    StoreOperation.PUT,
                    StoreOperation.DELETE,
                    StoreOperation.GET,
                ],
                single_collection=True,
            )
        elif op_parser.op_equals(StoreOperation.TRANSACT):
            Validator.validate_transact(
                op_parser.get_operation_parsers(),
                allowed_ops=[
                    StoreOperation.PUT,
                    StoreOperation.DELETE,
                    StoreOperation.GET,
                ],
                single_collection=True,
            )

    def run(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
    ) -> Any:
        self.init(context=context)
        op_parser = self.get_op_parser(operation)
        self._validate(op_parser)
        collection = self._get_collection(op_parser)
        ncall, state = self._get_ncall(
            op_parser,
            collection,
            ResourceHelper(self._database_client),
        )
        if ncall is None:
            return super().run(operation)
        nresult = ncall.invoke()
        result = self._convert_nresult(nresult, state, op_parser, collection)
        return Response(result=result, native=dict(result=nresult, call=ncall))

    def _get_ncall(
        self,
        op_parser: StoreOperationParser,
        collection: WeaviateCollection | None,
        resource_helper: Any,
    ) -> tuple[NCall | None, dict | None]:
        if collection is not None:
            converter = collection.converter
            client = collection.client
            helper = collection.helper
        call = None
        state = None
        nargs = op_parser.get_nargs()
        # CREATE COLLECTION
        if op_parser.op_equals(StoreOperation.CREATE_COLLECTION):
            args: dict[Any, Any] | None = {
                "collection": self._get_collection_name(op_parser),
                "config": get_collection_config(op_parser),
                "nargs": nargs,
            }
            call = NCall(resource_helper.create_collection, args)
        # DROP COLLECTION
        elif op_parser.op_equals(StoreOperation.DROP_COLLECTION):
            args = {
                "collection": self._get_collection_name(op_parser),
                "nargs": nargs,
            }
            call = NCall(resource_helper.drop_collection, args)
        # LIST COLLECTION
        elif op_parser.op_equals(StoreOperation.LIST_COLLECTIONS):
            args = {"nargs": nargs}
            call = NCall(resource_helper.list_collections, args)
        # GET
        elif op_parser.op_equals(StoreOperation.GET):
            args = converter.convert_get(op_parser.get_id_as_str())
            call = NCall(client.query.fetch_object_by_id, args, nargs)
        # PUT
        elif op_parser.op_equals(StoreOperation.PUT):
            args = {
                "id": op_parser.get_id_as_str(),
                "value": op_parser.get_vector_value(),
                "metadata": op_parser.get_metadata(),
                "nargs": nargs,
            }
            call = NCall(
                helper.put,
                args,
                None,
            )
        # UPDATE value
        elif (
            op_parser.op_equals(StoreOperation.UPDATE)
            and op_parser.is_value_update()
        ):
            args = converter.convert_update_value(
                id=op_parser.get_id_as_str(),
                value=op_parser.get_updated_value(),
            )
            call = NCall(client.data.update, args, nargs)
        # UPDATE metadata
        elif (
            op_parser.op_equals(StoreOperation.UPDATE)
            and op_parser.is_metadata_update()
        ):
            args = converter.convert_update_metadata(
                id=op_parser.get_id_as_str(),
                metadata=op_parser.get_updated_metadata(),
            )
            call = NCall(client.data.update, args, nargs)
        # DELETE
        elif op_parser.op_equals(StoreOperation.DELETE):
            args, func = converter.convert_delete(
                id=op_parser.get_id_as_str_or_none(),
                where=op_parser.get_where(),
            )
            if func == "delete_by_id":
                call = NCall(client.data.delete_by_id, args, nargs)
            elif func == "delete_many":
                call = NCall(client.data.delete_many, args, nargs)
        # QUERY
        elif op_parser.op_equals(StoreOperation.QUERY):
            args = converter.convert_query(
                search=op_parser.get_search_as_function(),
                where=op_parser.get_where(),
                select=op_parser.get_select(),
                order_by=op_parser.get_order_by(),
                limit=op_parser.get_limit(),
                offset=op_parser.get_offset(),
            )
            call = NCall(client.query.near_vector, args, nargs)
        # COUNT
        elif op_parser.op_equals(StoreOperation.COUNT):
            args = converter.convert_count(
                where=op_parser.get_where(),
            )
            call = NCall(client.aggregate.over_all, args, nargs)
        # BATCH or TRANSACT
        elif op_parser.op_equals(
            StoreOperation.BATCH
        ) or op_parser.op_equals(StoreOperation.TRANSACT):
            args, func = converter.convert_batch(
                op_parser.get_operation_parsers()
            )
            if func == "get":
                call = NCall(client.query.fetch_objects, args, nargs)
            elif func == "put":
                call = NCall(client.data.insert_many, args, nargs)
            elif func == "delete":
                call = NCall(client.data.delete_many, args, nargs)
            state = {"func": func}
        # CLOSE
        elif op_parser.op_equals(StoreOperation.CLOSE):
            args = {"nargs": nargs}
            call = NCall(resource_helper.close, args)
        return call, state

    def _convert_nresult(
        self,
        nresult: Any,
        state: dict | None,
        op_parser: StoreOperationParser,
        collection: WeaviateCollection | None,
    ) -> Any:
        if collection is not None:
            processor = collection.processor
        result: Any = None
        # CREATE COLLECTION
        if op_parser.op_equals(StoreOperation.CREATE_COLLECTION):
            result = None
        # DROP COLLECTION
        elif op_parser.op_equals(StoreOperation.DROP_COLLECTION):
            result = None
        # LIST COLLECTIONS
        elif op_parser.op_equals(StoreOperation.LIST_COLLECTIONS):
            result = [col.lower() for col in list(nresult.keys())]
        # GET
        elif op_parser.op_equals(StoreOperation.GET):
            if nresult is None:
                raise NotFoundError
            result = self._convert_to_item(nresult, processor)
        # PUT
        elif op_parser.op_equals(StoreOperation.PUT):
            result = VectorItem(
                key=VectorKey(id=op_parser.get_id_as_str()),
            )
        # UPDATE value
        elif (
            op_parser.op_equals(StoreOperation.UPDATE)
            and op_parser.is_value_update()
        ):
            result = VectorItem(key=VectorKey(id=op_parser.get_id_as_str()))
        # UPDATE metadata
        elif (
            op_parser.op_equals(StoreOperation.UPDATE)
            and op_parser.is_metadata_update()
        ):
            result = VectorItem(
                key=VectorKey(id=op_parser.get_id_as_str()),
                metadata=op_parser.get_updated_metadata(),
            )
        # DELETE
        elif op_parser.op_equals(StoreOperation.DELETE):
            result = None
        # QUERY
        elif op_parser.op_equals(StoreOperation.QUERY):
            items = []
            for item in nresult.objects:
                items.append(self._convert_to_item(item, processor))
            items = QueryProcessor.order_items(
                items, op_parser.get_order_by(), processor.resolve_root_field
            )
            result = VectorList(items=items)
        # COUNT
        elif op_parser.op_equals(StoreOperation.COUNT):
            result = nresult.total_count
        # BATCH or TRANSACT
        elif op_parser.op_equals(
            StoreOperation.BATCH
        ) or op_parser.op_equals(StoreOperation.TRANSACT):
            result = []
            func = state["func"] if state is not None else None
            if func == "get":
                for item in nresult.objects:
                    result.append(self._convert_to_item(item, processor))
            elif func == "put":
                for op_parser in op_parser.get_operation_parsers():
                    result.append(
                        VectorItem(key=VectorKey(id=op_parser.get_id_as_str()))
                    )
            elif func == "delete":
                for op_parser in op_parser.get_operation_parsers():
                    result.append(None)
        return result

    def _convert_to_item(
        self, nitem: Any, processor: ItemProcessor
    ) -> VectorItem:
        content = None
        metadata: dict | None = None
        vector = None
        key = VectorKey(id=str(nitem.uuid))
        if nitem.properties is not None:
            metadata = nitem.properties
            if (
                metadata is not None
                and self.content_properties_field is not None
                and self.content_properties_field in metadata
            ):
                content = metadata[self.content_properties_field]
                metadata.pop(self.content_properties_field)
                if len(metadata) == 0:
                    metadata = None
        if nitem.vector is not None and len(nitem.vector) > 0:
            vector = nitem.vector["default"]
        value = VectorValue(
            vector=vector,
            sparse_vector=None,
            content=content,
        )
        properties = None
        if (
            nitem.metadata is not None
            and hasattr(nitem.metadata, "distance")
            and nitem.metadata.distance is not None
        ):
            properties = VectorProperties(score=nitem.metadata.distance * -1)
        return VectorItem(
            key=key, value=value, metadata=metadata, properties=properties
        )


class ResourceHelper:
    database_client: weaviate.WeaviateClient

    def __init__(self, database_client: weaviate.WeaviateClient):
        self.database_client = database_client

    def create_collection(
        self,
        collection: str,
        config: VectorCollectionConfig | None,
        nargs: Any,
    ):
        metric_mapping = {
            VectorIndexMetric.DOTPRODUCT: wvc.config.VectorDistances.DOT,
            VectorIndexMetric.COSINE: wvc.config.VectorDistances.COSINE,
            VectorIndexMetric.EUCLIDEAN: wvc.config.VectorDistances.L2_SQUARED,
            VectorIndexMetric.MANHATTAN: wvc.config.VectorDistances.MANHATTAN,
            VectorIndexMetric.HAMMING: wvc.config.VectorDistances.HAMMING,
        }
        vector_index = config.vector_index if config is not None else None
        if vector_index is not None:
            metric = metric_mapping[vector_index.metric]
            structure = vector_index.structure
            nconfig = (
                vector_index.nconfig
                if vector_index.nconfig is not None
                else {}
            )
        else:
            metric = wvc.config.VectorDistances.DOT
            structure = None
            nconfig = {}
        if structure is None or structure == VectorIndexStructure.HNSW:
            vector_index_config: Any = wvc.config.Configure.VectorIndex.hnsw(
                distance_metric=metric, **nconfig
            )
        elif structure == VectorIndexStructure.FLAT:
            vector_index_config = wvc.config.Configure.VectorIndex.flat(
                distance_metric=metric, **nconfig
            )
        args = {
            "name": collection,
            "vectorizer_config": wvc.config.Configure.Vectorizer.none(),
            "vector_index_config": vector_index_config,
        }
        try:
            return NCall(
                self.database_client.collections.create,
                args,
                nargs,
                {weaviate.exceptions.UnexpectedStatusCodeError: None},
            ).invoke()
        except weaviate.exceptions.UnexpectedStatusCodeError as e:
            if e.status_code == 422:
                return None
            raise e

    def drop_collection(self, collection: str, nargs: Any):
        args = {"name": collection}
        return NCall(
            self.database_client.collections.delete, args, nargs
        ).invoke()

    def list_collections(self, nargs) -> Any:
        return NCall(
            self.database_client.collections.list_all, None, nargs
        ).invoke()

    def close(self, nargs: Any) -> Any:
        return NCall(self.database_client.close, None, nargs).invoke()


class ClientHelper:
    client: Any
    converter: OperationConverter

    def __init__(self, client: Any, converter: OperationConverter) -> None:
        self.client = client
        self.converter = converter

    def put(self, id: str, value: dict, metadata: dict, nargs: Any) -> Any:
        args = self.converter.convert_put(id, value, metadata)
        try:
            return NCall(self.client.data.insert, args, nargs).invoke()
        except weaviate.exceptions.UnexpectedStatusCodeError as e:
            if e.status_code == 422:
                return NCall(self.client.data.replace, args, nargs).invoke()
            raise e

    def delete(self, id: str, where: Expression, nargs: Any) -> Any:
        if id is None and where is None:
            raise BadRequestError("Id or where should be provided for delete")
        if id is not None and where is None:
            args = {"uuid": id}
            return NCall(self.client.data.delete_by_id, args, nargs).invoke()
        if where is not None and id is None:
            args = {"where": self.converter.convert_expr(where)}
            return NCall(self.client.data.delete_many, args, nargs).invoke()
        raise BadRequestError(
            "Both id and where cannot be provided for delete"
        )


class OperationConverter:
    processor: ItemProcessor
    collection: str | None
    dimension: int
    content_properties_field: str | None

    def __init__(
        self,
        processor: ItemProcessor,
        collection: str | None,
        dimension: int,
        content_properties_field: str | None,
    ) -> None:
        self.processor = processor
        self.collection = collection
        self.dimension = dimension
        self.content_properties_field = content_properties_field

    def convert_batch(
        self, op_parsers: list[StoreOperationParser]
    ) -> tuple[dict, str]:
        batch_get = False
        batch_put = False
        batch_delete = False
        for op_parser in op_parsers:
            if op_parser.op_equals(StoreOperation.GET):
                batch_get = True
            elif op_parser.op_equals(StoreOperation.PUT):
                batch_put = True
            elif op_parser.op_equals(StoreOperation.DELETE):
                batch_delete = True
        if (
            (batch_get and batch_put)
            or (batch_put and batch_delete)
            or (batch_delete and batch_get)
        ):
            raise BadRequestError(
                "Batch operation supports only one operation type"
            )
        if batch_get:
            return self.convert_batch_get(op_parsers), "get"
        elif batch_put:
            return self.convert_batch_put(op_parsers), "put"
        elif batch_delete:
            return self.convert_batch_delete(op_parsers), "delete"
        raise BadRequestError("Batch operation not supported")

    def convert_batch_get(
        self, op_parsers: list[StoreOperationParser]
    ) -> dict:
        ids: list = []
        for op_parser in op_parsers:
            ids.append(op_parser.get_id_as_str())
        return {
            "filters": Filter.by_id().contains_any(ids),
            "include_vector": True,
        }

    def convert_batch_put(
        self, op_parsers: list[StoreOperationParser]
    ) -> dict:
        objs = list()
        for op_parser in op_parsers:
            objs.append(
                wvc.data.DataObject(
                    **self.convert_put(
                        op_parser.get_id_as_str(),
                        op_parser.get_vector_value(),
                        op_parser.get_metadata(),
                    )
                )
            )
        return {"objects": objs}

    def convert_batch_delete(
        self, op_parsers: list[StoreOperationParser]
    ) -> dict:
        ids: list = []
        for op_parser in op_parsers:
            ids.append(op_parser.get_id_as_str())
        return {"where": Filter.by_id().contains_any(ids)}

    def convert_get(self, id: str) -> dict:
        return {"uuid": id, "include_vector": True}

    def convert_put(self, id: str, value: dict, metadata: dict | None) -> dict:
        args: dict = {"uuid": id}
        content = None
        md = None
        vector_value = VectorValue(**value)
        if vector_value.vector is not None:
            args["vector"] = vector_value.vector

        content = vector_value.content
        if metadata is None:
            if (
                content is not None
                and self.content_properties_field is not None
            ):
                md = {self.content_properties_field: content}
        else:
            md = copy.deepcopy(metadata)
            if (
                content is not None
                and self.content_properties_field is not None
            ):
                md[self.content_properties_field] = content
        if md is not None:
            args["properties"] = md
        return args

    def convert_update_value(self, id: str, value: dict) -> dict:
        args: dict = {"uuid": id}
        vector_value = VectorValue(**value)
        if vector_value.vector is not None:
            args["vector"] = vector_value.vector
        if (
            vector_value.content is not None
            and self.content_properties_field is not None
        ):
            args["properties"] = {
                self.content_properties_field: vector_value.content
            }
        return args

    def convert_update_metadata(self, id: str, metadata: dict | None) -> dict:
        return {
            "uuid": id,
            "properties": metadata,
        }

    def convert_delete(self, id: str | None, where: Expression) -> Any:
        if id is None and where is None:
            raise BadRequestError("Id or where should be provided for delete")
        if id is not None and where is None:
            return {"uuid": id}, "delete_by_id"
        if where is not None and id is None:
            return {"where": self.convert_expr(where)}, "delete_many"
        raise BadRequestError(
            "Both id and where cannot be provided for delete"
        )

    def convert_query(
        self,
        search: Function | None,
        where: Expression,
        select: Select | None,
        order_by: OrderBy | None,
        limit: int | None,
        offset: int | None,
    ) -> dict:
        args: dict = dict()
        search_args = self.convert_search(search)
        args = args | search_args
        select_args = self.convert_select(select)
        if select_args is not None:
            args = args | select_args
        if where is not None:
            args["filters"] = self.convert_expr(where)
        args["limit"] = limit
        args["offset"] = offset
        return args

    def convert_select(self, select: Select | None) -> dict:
        include_value = False
        if select is None or len(select.terms) == 0:
            include_value = True
        else:
            for term in select.terms:
                if is_value_field(term.field):
                    include_value = True
                elif is_metadata_field(term.field):
                    pass
        return {
            "include_vector": include_value,
            "return_metadata": wvc.query.MetadataQuery(
                distance=True, score=True
            ),
        }

    def convert_search(self, search: Function | None) -> dict:
        if search is None:
            return {"near_vector": [0.0] * self.dimension}
        namespace = search.namespace
        name = search.name
        if namespace == FunctionNamespace.BUILTIN:
            if name == StoreFunctionName.VECTOR_SEARCH:
                args = VectorSearchArgs(**search.named_args)
                return {
                    "near_vector": args.vector,
                }

        raise BadRequestError("Search function not supported")

    def convert_expr(self, expr: Expression) -> Any:
        if expr is None or isinstance(
            expr, (str, int, float, bool, dict, list)
        ):
            return expr
        if isinstance(expr, Field):
            field = self.processor.resolve_metadata_field(expr.path)
            return field
        if isinstance(expr, Function):
            return self.convert_func(expr)
        if isinstance(expr, Comparison):
            op = expr.op
            if isinstance(expr.lexpr, Field):
                field = self.convert_expr(expr.lexpr)
                value = self.convert_expr(expr.rexpr)
            elif isinstance(expr.rexpr, Field):
                field = self.convert_expr(expr.rexpr)
                value = self.convert_expr(expr.lexpr)
                op = Comparison.reverse_op(op)

            if op == ComparisonOp.EQ:
                return Filter.by_property(field).equal(value)
            elif op == ComparisonOp.NEQ:
                return Filter.by_property(field).not_equal(value)
            elif op == ComparisonOp.GT:
                return Filter.by_property(field).greater_than(value)
            elif op == ComparisonOp.GTE:
                return Filter.by_property(field).greater_or_equal(value)
            elif op == ComparisonOp.LT:
                return Filter.by_property(field).less_than(value)
            elif op == ComparisonOp.LTE:
                return Filter.by_property(field).less_or_equal(value)
            elif op == ComparisonOp.BETWEEN:
                return Filter.by_property(field).greater_or_equal(
                    value[0]
                ) & Filter.by_property(field).less_or_equal(value[1])
            elif op == ComparisonOp.IN:
                x = Filter.by_property(field).equal(value[0])
                for i in range(0, len(value)):
                    x = x | Filter.by_property(field).equal(value[i])
                return x
            elif op == ComparisonOp.NIN:
                x = Filter.by_property(field).not_equal(value[0])
                for i in range(0, len(value)):
                    x = x & Filter.by_property(field).not_equal(value[i])
                return x
            elif op == ComparisonOp.LIKE:
                return Filter.by_property(field).like(value)
            else:
                raise BadRequestError(f"Comparison op {expr.op} not supported")
        if isinstance(expr, And):
            return self.convert_expr(expr.lexpr) & self.convert_expr(
                expr.rexpr
            )
        if isinstance(expr, Or):
            return self.convert_expr(expr.lexpr) | self.convert_expr(
                expr.rexpr
            )
        if isinstance(expr, Not):
            raise BadRequestError("Not expression is not supported")
        return str(expr)

    def convert_func(self, expr: Function) -> Any:
        namespace = expr.namespace
        name = expr.name
        args = expr.args
        if namespace == FunctionNamespace.BUILTIN:
            if name == StoreFunctionName.IS_DEFINED:
                field = self.convert_expr(args[0])
                return Filter.by_property(field).is_none(False)
            if name == StoreFunctionName.IS_NOT_DEFINED:
                field = self.convert_expr(args[0])
                return Filter.by_property(field).is_none(True)
            if name == StoreFunctionName.ARRAY_CONTAINS:
                field = self.convert_expr(args[0])
                return Filter.by_property(field).contains_all([args[1]])
            if name == StoreFunctionName.ARRAY_CONTAINS_ANY:
                field = self.convert_expr(args[0])
                return Filter.by_property(field).contains_any(args[1])
        raise BadRequestError(f"Function {name} not supported")

    def convert_count(self, where: Expression) -> dict | None:
        if where is None:
            return None
        args = {"filters": self.convert_expr(where)}
        return args


class WeaviateCollection:
    client: Any
    converter: OperationConverter
    processor: ItemProcessor
    helper: ClientHelper

    def __init__(
        self,
        client: Any,
        collection: str | None,
        dimension: int,
        content_properties_field: str | None,
    ):
        self.client = client
        self.processor = ItemProcessor()
        self.converter = OperationConverter(
            processor=self.processor,
            collection=collection,
            dimension=dimension,
            content_properties_field=content_properties_field,
        )
        self.helper = ClientHelper(self.client, self.converter)



"""
Object Store on Amazon S3.
"""

from __future__ import annotations

__all__ = ["AmazonS3"]

import json
from io import BytesIO
from typing import IO, Any

import boto3
import s3transfer.tasks
import s3transfer.upload
from boto3.s3.transfer import TransferConfig as S3TransferConfig
from botocore.exceptions import ClientError

from verse.core import Context, NCall, Operation, Response
from verse.core.exceptions import (
    BadRequestError,
    NotFoundError,
    NotModified,
    PreconditionFailedError,
)
from verse.internal.storage_core import (
    ItemProcessor,
    MatchCondition,
    StoreOperation,
    StoreOperationParser,
    StoreProvider,
)

from .._helper import (
    QueryArgs,
    get_collection_config,
    get_query_args,
    get_transfer_config,
)
from .._models import (
    ObjectCollectionConfig,
    ObjectItem,
    ObjectKey,
    ObjectList,
    ObjectProperties,
    ObjectStoreClass,
    ObjectTransferConfig,
    ObjectVersion,
)


# Monkey patching to return Etag and Version from upload* methods.
# https://github.com/boto/s3transfer/issues/82#issuecomment-837971614
# -----------------------------
class PutObjectTask(s3transfer.tasks.Task):
    # Copied from s3transfer/upload.py,
    # changed to return the result of client.put_object.
    def _main(self, client, fileobj, bucket, key, extra_args):
        with fileobj as body:
            return client.put_object(
                Bucket=bucket, Key=key, Body=body, **extra_args
            )


class CompleteMultipartUploadTask(s3transfer.tasks.Task):
    # Copied from s3transfer/tasks.py, changed to return a result.
    def _main(self, client, bucket, key, upload_id, parts, extra_args):
        return client.complete_multipart_upload(
            Bucket=bucket,
            Key=key,
            UploadId=upload_id,
            MultipartUpload={"Parts": parts},
            **extra_args,
        )


s3transfer.upload.PutObjectTask = PutObjectTask
s3transfer.upload.CompleteMultipartUploadTask = CompleteMultipartUploadTask
# -----------------------------


class AmazonS3(StoreProvider):
    region_name: str | None
    profile_name: str | None
    aws_access_key_id: str | None
    aws_secret_access_key: str | None
    aws_session_token: str | None
    bucket: str | None
    nparams: dict[str, Any]

    _resource: Any

    _collection_cache: dict[str, AmazonS3Collection]

    def __init__(
        self,
        region_name: str | None = None,
        profile_name: str | None = None,
        aws_access_key_id: str | None = None,
        aws_secret_access_key: str | None = None,
        aws_session_token: str | None = None,
        bucket: str | None = None,
        nparams: dict[str, Any] = dict(),
        **kwargs,
    ):
        """Initialize.

        Args:
            region_name:
                AWS region name.
            profile_name:
                AWS profile name.
            aws_access_key_id:
                AWS access key id.
            aws_secret_access_key:
                AWS secret access key.
            aws_session_token:
                AWS session token.
            bucket:
                S3 bucket mapped to object store collection.
            nparams:
                Native parameters to boto3 client.
        """
        self.region_name = region_name
        self.profile_name = profile_name
        self.aws_access_key_id = aws_access_key_id
        self.aws_secret_access_key = aws_secret_access_key
        self.aws_session_token = aws_session_token
        self.bucket = bucket
        self.nparams = nparams

        self._resource = None
        self._collection_cache = dict()

    def init(self, context: Context | None = None) -> None:
        if self._resource is not None:
            return

        resource = None
        if self.profile_name is not None:
            resource = boto3.resource(
                "s3",
                region_name=self.region_name,
                profile_name=self.profile_name,
                **self.nparams,
            )
        elif (
            self.aws_access_key_id is not None
            and self.aws_secret_access_key is not None
        ):
            resource = boto3.resource(
                "s3",
                region_name=self.region_name,
                aws_access_key_id=self.aws_access_key_id,
                aws_secret_access_key=self.aws_secret_access_key,
                aws_session_token=self.aws_session_token,
                **self.nparams,
            )
        else:
            resource = boto3.resource(
                "s3",
                region_name=self.region_name,
                **self.nparams,
            )

        self._resource = resource

    def _get_bucket_name(self, op_parser: StoreOperationParser) -> str | None:
        collection_name = (
            op_parser.get_operation_parsers()[0].get_collection_name()
            if op_parser.op_equals(StoreOperation.BATCH)
            else op_parser.get_collection_name()
        )
        bucket = (
            collection_name if collection_name is not None else self.bucket
        )
        return bucket

    def _get_collection(
        self, op_parser: StoreOperationParser
    ) -> AmazonS3Collection | None:
        if op_parser.is_resource_op():
            return None
        bucket_name = self._get_bucket_name(op_parser)
        if bucket_name is None:
            raise BadRequestError("Collection name must be specified")
        if bucket_name in self._collection_cache:
            return self._collection_cache[bucket_name]
        client = self._resource.meta.client
        col = AmazonS3Collection(
            self._resource,
            client,
            bucket_name,
        )
        self._collection_cache[bucket_name] = col
        return col

    def _validate(self, op_parser: StoreOperationParser):
        pass

    def run(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
    ) -> Any:
        self.init(context=context)
        op_parser = self.get_op_parser(operation)
        self._validate(op_parser)
        collection = self._get_collection(op_parser)
        ncall, state = self._get_ncall(
            op_parser,
            collection,
            ResourceHelper(self._resource),
        )
        if ncall is None:
            return super().run(operation=operation, context=context)
        nresult, nerror = ncall.invoke(return_error=True)
        result = self._convert_nresult(
            nresult, nerror, state, op_parser, collection
        )
        return Response(result=result, native=dict(result=nresult, call=ncall))

    def _get_ncall(
        self,
        op_parser: StoreOperationParser,
        collection: AmazonS3Collection | None,
        resource_helper: Any,
    ) -> tuple[NCall | None, dict]:
        if collection is not None:
            op_converter = collection.op_converter
            client = collection.client
            helper = collection.helper
        call = None
        state: dict = {}
        nargs = op_parser.get_nargs()
        # CREATE COLLECTION
        if op_parser.op_equals(StoreOperation.CREATE_COLLECTION):
            args = {
                "bucket": self._get_bucket_name(op_parser),
                "config": get_collection_config(op_parser),
                "nargs": nargs,
            }
            call = NCall(resource_helper.create_collection, args)
        # DROP COLLECTION
        elif op_parser.op_equals(StoreOperation.DROP_COLLECTION):
            args = {
                "bucket": self._get_bucket_name(op_parser),
                "nargs": nargs,
            }
            call = NCall(resource_helper.drop_collection, args)
        # LIST COLLECTIONS
        elif op_parser.op_equals(StoreOperation.LIST_COLLECTIONS):
            args = {"nargs": nargs}
            call = NCall(resource_helper.list_collections, args)
        # PUT
        elif op_parser.op_equals(StoreOperation.PUT):
            args, func, state = op_converter.convert_put(
                id=op_parser.get_id_as_str(),
                value=op_parser.get_value_as_bytes(),
                file=op_parser.get_file(),
                stream=op_parser.get_stream(),
                metadata=op_parser.get_metadata(),
                properties=op_parser.get_properties(),
                match_condition=op_parser.get_match_condition(),
                config=get_transfer_config(op_parser),
            )
            if func == "put_object":
                call = NCall(client.put_object, args, nargs)
            elif func == "upload_fileobj":
                call = NCall(client.upload_fileobj, args, nargs)
            elif func == "upload_file":
                call = NCall(
                    helper.put,
                    {
                        "id": op_parser.get_id_as_str(),
                        "args": args,
                        "returning": op_parser.get_returning(),
                        "nargs": nargs,
                    },
                )
        # GET value
        elif (
            op_parser.op_equals(StoreOperation.GET)
            and op_parser.is_value_attribute()
        ):
            args, func, state = op_converter.convert_get(
                id=op_parser.get_id_as_str(),
                version=op_parser.get_version(),
                file=op_parser.get_file(),
                stream=op_parser.get_stream(),
                match_condition=op_parser.get_match_condition(),
                start=op_parser.get_start(),
                end=op_parser.get_end(),
                config=get_transfer_config(op_parser),
            )
            if func == "get_object":
                call = NCall(client.get_object, args, nargs)
            elif func == "download_fileobj":
                call = NCall(client.download_fileobj, args, nargs)
            elif func == "download_file":
                call = NCall(client.download_file, args, nargs)
        # GET metadata or properties
        elif op_parser.op_equals(StoreOperation.GET) and (
            op_parser.is_metadata_attribute()
            or op_parser.is_properties_attribute()
        ):
            args = op_converter.convert_get_properties(
                id=op_parser.get_id_as_str(),
                version=op_parser.get_version(),
                match_condition=op_parser.get_match_condition(),
            )
            call = NCall(client.head_object, args, nargs)
        # GET versions
        elif op_parser.op_equals(StoreOperation.GET) and (
            op_parser.is_versions_attribute()
        ):
            args = op_converter.convert_get_versions(
                id=op_parser.get_id_as_str()
            )
            call = NCall(client.list_object_versions, args, nargs)
        # UPDATE
        elif op_parser.op_equals(StoreOperation.UPDATE):
            args = op_converter.convert_update(
                id=op_parser.get_id_as_str(),
                version=op_parser.get_version(),
                metadata=op_parser.get_updated_metadata(),
                properties=op_parser.get_updated_properities(),
                match_condition=op_parser.get_match_condition(),
            )
            call = NCall(client.copy_object, args, nargs)
        # DELETE
        elif op_parser.op_equals(StoreOperation.DELETE):
            args, func = op_converter.convert_delete(
                id=op_parser.get_id_as_str(), version=op_parser.get_version()
            )
            if func == "helper_delete":
                args["nargs"] = nargs
                call = NCall(helper.delete, args, None)
            elif func == "delete_object":
                call = NCall(client.delete_object, args, nargs)
        # COPY
        elif op_parser.op_equals(StoreOperation.COPY):
            args = op_converter.convert_copy(
                id=op_parser.get_id_as_str(),
                source_id=op_parser.get_source_id_as_str(),
                source_version=op_parser.get_source_version(),
                source_collection=op_parser.get_source_collection(),
                metadata=op_parser.get_metadata(),
                properties=op_parser.get_properties(),
                match_condition=op_parser.get_match_condition(),
            )
            call = NCall(client.copy_object, args, nargs)
        # GENERATE signed url
        elif op_parser.op_equals(StoreOperation.GENERATE):
            args = op_converter.convert_generate(
                id=op_parser.get_id_as_str(),
                version=op_parser.get_version(),
                method=op_parser.get_method(),
                expiry=op_parser.get_expiry(),
            )
            call = NCall(client.generate_presigned_url, args, nargs)
        # QUERY
        elif op_parser.op_equals(StoreOperation.QUERY):
            query_args = get_query_args(op_parser)
            state["query_args"] = query_args
            args = {"query_args": get_query_args(op_parser), "nargs": nargs}
            call = NCall(helper.query, args, None)
        # COUNT
        elif op_parser.op_equals(StoreOperation.COUNT):
            query_args = get_query_args(op_parser)
            state["query_args"] = query_args
            args = {"query_args": get_query_args(op_parser), "nargs": nargs}
            call = NCall(helper.query, args, None)
        # BATCH
        elif op_parser.op_equals(StoreOperation.BATCH):
            args, func = op_converter.convert_batch(
                op_parser.get_operation_parsers()
            )
            if func == "delete":
                call = NCall(client.delete_objects, args, nargs)
            state["func"] = func
        # CLOSE
        elif op_parser.op_equals(StoreOperation.CLOSE):
            args = {"nargs": nargs}
            call = NCall(resource_helper.close, args)
        return call, state

    def _convert_nresult(
        self,
        nresult: Any,
        nerror: Any,
        state: dict,
        op_parser: StoreOperationParser,
        collection: AmazonS3Collection | None,
    ) -> Any:
        if collection is not None:
            result_converter = collection.result_converter
        result: Any = None

        def check_error(nerror):
            if nerror is not None:
                if isinstance(nerror, ClientError):
                    if (
                        nerror.response["Error"]["Code"] == "NoSuchKey"
                        or nerror.response["Error"]["Code"] == "404"
                    ):
                        raise NotFoundError
                    elif (
                        nerror.response["Error"]["Code"]
                        == "PreconditionFailed"
                        or nerror.response["Error"]["Code"] == "412"
                    ):
                        raise PreconditionFailedError
                    elif nerror.response["Error"]["Code"] == "304":
                        raise NotModified
                raise nerror

        # CREATE COLLECTION
        if op_parser.op_equals(StoreOperation.CREATE_COLLECTION):
            if nerror is not None:
                if (
                    isinstance(nerror, ClientError)
                    and nerror.response["Error"]["Code"]
                    == "ResourceInUseException"
                ):
                    pass
                else:
                    raise nerror
            result = None
        # DROP COLLECTION
        elif op_parser.op_equals(StoreOperation.DROP_COLLECTION):
            if nerror is not None:
                raise nerror
            result = None
        # LIST COLLECTIONS
        elif op_parser.op_equals(StoreOperation.LIST_COLLECTIONS):
            if nerror is not None:
                raise nerror
            result = [b.name for b in nresult]
        # PUT
        elif op_parser.op_equals(StoreOperation.PUT):
            if nerror is not None:
                raise nerror
            if "open_file" in state:
                state["open_file"].close()
            result = result_converter.convert_put_result(
                op_parser.get_id_as_str(), nresult
            )
        # GET value
        elif (
            op_parser.op_equals(StoreOperation.GET)
            and op_parser.is_value_attribute()
        ):
            check_error(nerror)
            result = result_converter.convert_get_result(
                op_parser.get_id_as_str(),
                op_parser.get_file(),
                op_parser.get_stream(),
                nresult,
                state,
            )
        # GET metadata or properties
        elif op_parser.op_equals(StoreOperation.GET) and (
            op_parser.is_metadata_attribute()
            or op_parser.is_properties_attribute()
        ):
            check_error(nerror)
            result = result_converter.convert_get_properties_result(
                op_parser.get_id_as_str(), nresult
            )
        # GET versions
        elif op_parser.op_equals(StoreOperation.GET) and (
            op_parser.is_versions_attribute()
        ):
            check_error(nerror)
            result = result_converter.convert_get_versions_result(
                op_parser.get_id_as_str(), nresult
            )
        # UPDATE
        elif op_parser.op_equals(StoreOperation.UPDATE):
            check_error(nerror)
            result = result_converter.convert_copy_result(
                op_parser.get_id_as_str(), nresult
            )
        # DELETE
        elif op_parser.op_equals(StoreOperation.DELETE):
            check_error(nerror)
            result = None
        # COPY
        elif op_parser.op_equals(StoreOperation.COPY):
            check_error(nerror)
            result = result_converter.convert_copy_result(
                op_parser.get_id_as_str(), nresult
            )
        # GENERATE signed url
        elif op_parser.op_equals(StoreOperation.GENERATE):
            check_error(nerror)
            result = result_converter.convert_generate_result(
                op_parser.get_id_as_str(),
                op_parser.get_version(),
                nresult,
            )
        # QUERY
        elif op_parser.op_equals(StoreOperation.QUERY):
            if nerror is not None:
                raise nerror
            result = result_converter.convert_query_result(
                nresult,
                state["query_args"],
            )
        # COUNT
        elif op_parser.op_equals(StoreOperation.COUNT):
            if nerror is not None:
                raise nerror
            result = result_converter.convert_count_result(
                nresult,
            )
        # BATCH
        elif op_parser.op_equals(StoreOperation.BATCH):
            result = []
            func = state["func"]
            if func == "delete":
                for op_parser in op_parser.get_operation_parsers():
                    result.append(None)
        return result


class ResourceHelper:
    resource: Any

    def __init__(self, resource: Any):
        self.resource = resource

    def create_collection(
        self, bucket: str, config: ObjectCollectionConfig | None, nargs: Any
    ):
        args: dict = {
            "Bucket": bucket,
        }
        if config is not None and config.nconfig is not None:
            args = args | config.nconfig
        if "CreateBucketConfiguration" not in args:
            region_name = self.resource.meta.client.meta.region_name
            args["CreateBucketConfiguration"] = {
                "LocationConstraint": region_name
            }
        s3_bucket = NCall(
            self.resource.create_bucket,
            args,
            nargs,
        ).invoke()
        if config is not None and config.acl is not None:
            if config.acl == "public-read":
                NCall(
                    self.resource.meta.client.put_public_access_block,
                    {
                        "Bucket": bucket,
                        "PublicAccessBlockConfiguration": {
                            "BlockPublicPolicy": False,
                        },
                    },
                    None,
                ).invoke()
                policy = {
                    "Version": "2012-10-17",
                    "Statement": [
                        {
                            "Sid": "PublicReadGetObject",
                            "Effect": "Allow",
                            "Principal": "*",
                            "Action": ["s3:GetObject"],
                            "Resource": [f"arn:aws:s3:::{bucket}/*"],
                        }
                    ],
                }
                NCall(
                    self.resource.meta.client.put_bucket_policy,
                    {"Bucket": bucket, "Policy": json.dumps(policy)},
                    None,
                ).invoke()
        if config is not None and config.versioned:
            NCall(
                self.resource.meta.client.put_bucket_versioning,
                {
                    "Bucket": bucket,
                    "VersioningConfiguration": {"Status": "Enabled"},
                },
                None,
            ).invoke()
        NCall(s3_bucket.wait_until_exists, None, nargs).invoke()

    def drop_collection(self, bucket: str, nargs: Any):
        NCall(
            self.resource.meta.client.delete_bucket, {"Bucket": bucket}, nargs
        ).invoke()
        waiter = NCall(
            self.resource.meta.client.get_waiter, ["bucket_not_exists"], nargs
        ).invoke()
        NCall(waiter.wait, {"Bucket": bucket}, nargs).invoke()

    def list_collections(self, nargs) -> Any:
        response = NCall(self.resource.buckets.all, None, nargs).invoke()
        return response

    def close(self, nargs: Any) -> Any:
        pass


class ClientHelper:
    resource: Any
    client: Any
    bucket: str

    def __init__(self, resource: Any, client: Any, bucket: str) -> None:
        self.resource = resource
        self.client = client
        self.bucket = bucket

    def put(
        self, id: str, args: dict, returning: bool | None, nargs: Any
    ) -> Any:
        NCall(self.client.upload_file, args, nargs).invoke()
        if returning is True:
            args = {"Bucket": self.bucket, "Key": id}
            return NCall(self.client.head_object, args, None).invoke()
        return None

    def delete(self, id: str, version: str | None, nargs: Any) -> Any:
        if version == "*":
            return NCall(
                self.resource.Bucket(self.bucket)
                .object_versions.filter(Prefix=id)
                .delete,
                None,
                nargs,
            ).invoke()
        return None

    def query(
        self,
        query_args: QueryArgs,
        nargs: Any,
    ) -> Any:
        responses: list = []
        args: dict = {}
        args["Bucket"] = self.bucket
        if query_args.prefix is not None:
            args["Prefix"] = query_args.prefix
        if query_args.delimiter is not None:
            args["Delimiter"] = query_args.delimiter
        if query_args.start_after is not None:
            args["StartsAfter"] = query_args.start_after
        continuation_token = query_args.continuation
        total_key_count = 0
        while True:
            if continuation_token is not None:
                args["ContinuationToken"] = continuation_token
            if (
                query_args.limit is not None
                and query_args.page_size is not None
            ):
                if query_args.page_size < query_args.limit - total_key_count:
                    args["MaxKeys"] = query_args.page_size
                else:
                    args["MaxKeys"] = query_args.limit - total_key_count
            elif query_args.limit is not None:
                args["MaxKeys"] = query_args.limit - total_key_count
            elif query_args.page_size is not None:
                args["MaxKeys"] = query_args.page_size
            response = NCall(self.client.list_objects_v2, args, nargs).invoke()
            responses.append(response)
            key_count = response["KeyCount"]
            total_key_count += key_count
            if (
                query_args.limit is not None
                and total_key_count >= query_args.limit
            ):
                break
            if "NextContinuationToken" in response:
                continuation_token = response["NextContinuationToken"]
            else:
                break
            if query_args.paging:
                break
        return responses


class ResultConverter:
    processor: ItemProcessor
    bucket: str

    def __init__(self, processor: ItemProcessor, bucket: str):
        self.processor = processor
        self.bucket = bucket

    def convert_url(self, id: str):
        return f"https://{self.bucket}.s3.amazonaws.com/{id}"

    def convert_storage_class(self, storage_class: str | None) -> str:
        map = {
            "STANDARD": ObjectStoreClass.HOT,
            "STANDARD_IA": ObjectStoreClass.COOL,
            "GLACIER": ObjectStoreClass.COLD,
            "DEEP_ARCHIVE": ObjectStoreClass.ARCHIVE,
        }
        if storage_class is None:
            return ObjectStoreClass.HOT.value
        if storage_class in map:
            return map[storage_class].value
        return storage_class

    def convert_properties(self, nresult: Any) -> ObjectProperties:
        properties = ObjectProperties()
        if "CacheControl" in nresult:
            properties.cache_control = nresult["CacheControl"]
        if "ContentDisposition" in nresult:
            properties.content_disposition = nresult["ContentDisposition"]
        if "ContentEncoding" in nresult:
            properties.content_encoding = nresult["ContentEncoding"]
        if "ContentLanguage" in nresult:
            properties.content_language = nresult["ContentLanguage"]
        if "ContentLength" in nresult:
            properties.content_length = nresult["ContentLength"]
        if "ContentMD5" in nresult:
            properties.content_md5 = nresult["ContentMD5"]
        if "ContentType" in nresult:
            properties.content_type = nresult["ContentType"]
        if "ChecksumCRC32C" in nresult:
            properties.crc32c = nresult["ChecksumCRC32C"]
        if "LastModified" in nresult:
            properties.last_modified = nresult["LastModified"]
        if "ETag" in nresult:
            properties.etag = nresult["ETag"]
        if "Size" in nresult:
            properties.content_length = nresult["Size"]
        if "Expires" in nresult:
            properties.etag = nresult["Expires"]
        if "StorageClass" in nresult:
            properties.storage_class = self.convert_storage_class(
                nresult["StorageClass"]
            )
        else:
            properties.storage_class = ObjectStoreClass.HOT
        return properties

    def convert_put_result(self, id: str, nresult: Any) -> ObjectItem:
        version = None
        etag = None
        if nresult is not None and isinstance(nresult, dict):
            if "VersionId" in nresult:
                version = nresult["VersionId"]
            if "ETag" in nresult:
                etag = nresult["ETag"]
        return ObjectItem(
            key=ObjectKey(id=id, version=version),
            properties=ObjectProperties(etag=etag),
            url=self.convert_url(id=id),
        )

    def convert_get_result(
        self,
        id: str,
        file: str | None,
        stream: IO | None,
        nresult: Any,
        state: dict,
    ) -> ObjectItem:
        value = None
        version = None
        metadata = None
        func = state["func"]
        if func == "get_object":
            body = nresult["Body"]
            if file is not None:
                with open(file, "wb") as f:
                    f.write(body.read())
            elif stream is not None:
                stream.write(body.read())
            else:
                value = body.read()
            if "VersionId" in nresult:
                version = nresult["VersionId"]
            if "Metadata" in nresult:
                metadata = nresult["Metadata"]
            properties = self.convert_properties(nresult)
            return ObjectItem(
                key=ObjectKey(id=id, version=version),
                value=value,
                metadata=metadata,
                properties=properties,
                url=self.convert_url(id=id),
            )
        elif func == "download_file":
            return ObjectItem(
                key=ObjectKey(id=id), url=self.convert_url(id=id)
            )
        else:
            if stream is None:
                fileobj = state["fileobj"]
                fileobj.seek(0)
                value = fileobj.read()
                return ObjectItem(
                    key=ObjectKey(id=id),
                    value=value,
                    url=self.convert_url(id=id),
                )
            return ObjectItem(
                key=ObjectKey(id=id), url=self.convert_url(id=id)
            )

    def convert_get_properties_result(
        self, id: str, nresult: Any
    ) -> ObjectItem:
        version = None
        metadata = None
        properties = ObjectProperties()
        if "VersionId" in nresult:
            version = nresult["VersionId"]
        key = ObjectKey(id=id, version=version)
        if "Metadata" in nresult:
            metadata = nresult["Metadata"]
        properties = self.convert_properties(nresult)
        return ObjectItem(
            key=key,
            metadata=metadata,
            properties=properties,
            url=self.convert_url(id=id),
        )

    def convert_get_versions_result(self, id: str, nresult: Any) -> ObjectItem:
        key = ObjectKey(id=id)
        versions: list = []
        if "Versions" in nresult:
            for version in nresult["Versions"]:
                if version["Key"] != id:
                    continue
                properties = self.convert_properties(version)
                versions.append(
                    ObjectVersion(
                        version=version["VersionId"],
                        properties=properties,
                        latest=version["IsLatest"],
                    )
                )
        if len(versions) == 0:
            raise NotFoundError
        return ObjectItem(
            key=key, versions=versions[::-1], url=self.convert_url(id=id)
        )

    def convert_copy_result(self, id: str, nresult: Any) -> ObjectItem:
        version = None
        etag = None
        last_modified = None
        if "VersionId" in nresult:
            version = nresult["VersionId"]
        if "CopyObjectResult" in nresult:
            etag = nresult["CopyObjectResult"]["ETag"]
            last_modified = nresult["CopyObjectResult"]["LastModified"]
        return ObjectItem(
            key=ObjectKey(id=id, version=version),
            properties=ObjectProperties(
                etag=etag, last_modified=last_modified
            ),
            url=self.convert_url(id=id),
        )

    def convert_generate_result(
        self, id: str, version: str | None, nresult: Any
    ) -> ObjectItem:
        return ObjectItem(
            key=ObjectKey(id=id, version=version),
            url=nresult,
        )

    def convert_query_result(
        self, nresult: Any, query_args: QueryArgs
    ) -> ObjectList:
        items: list = []
        prefixes: list = []
        total_count = 0
        continuation = None
        for response in nresult:
            total_count += response["KeyCount"]
            if "Contents" in response:
                for content in response["Contents"]:
                    item = ObjectItem(
                        key=ObjectKey(id=content["Key"]),
                        properties=self.convert_properties(content),
                        url=self.convert_url(id=content["Key"]),
                    )
                    items.append(item)
            if "CommonPrefixes" in response:
                for prefix in response["CommonPrefixes"]:
                    prefixes.append(prefix["Prefix"])
            if "NextContinuationToken" in response:
                continuation = response["NextContinuationToken"]
            else:
                continuation = None
        if not query_args.paging:
            continuation = None
        if query_args.limit is not None:
            if total_count >= query_args.limit:
                continuation = None
        return ObjectList(
            items=items,
            continuation=continuation,
            prefixes=prefixes,
        )

    def convert_count_result(
        self,
        nresult: Any,
    ) -> int:
        total_count = 0
        for response in nresult:
            total_count += response["KeyCount"]
        return total_count


class OperationConverter:
    processor: ItemProcessor
    bucket: str

    def __init__(self, processor: ItemProcessor, bucket: str):
        self.processor = processor
        self.bucket = bucket

    def convert_transfer_config(
        self, config: ObjectTransferConfig | None
    ) -> S3TransferConfig:
        if config is None:
            return None
        args: dict = {}
        if config is not None:
            if config.multipart is not None:
                args["multipart_threshold"] = 1
            if config.concurrency is not None:
                args["max_concurrency"] = config.concurrency
            if config.chunksize is not None:
                args["multipart_chunksize"] = config.chunksize
            if config.nconfig is not None:
                args = args | config.nconfig
        return S3TransferConfig(**args)

    def convert_storage_class(self, storage_class: str | None) -> str:
        map = {
            ObjectStoreClass.HOT: "STANDARD",
            ObjectStoreClass.COOL: "STANDARD_IA",
            ObjectStoreClass.COLD: "GLACIER",
            ObjectStoreClass.ARCHIVE: "DEEP_ARCHIVE",
        }
        if storage_class is None:
            return map[ObjectStoreClass.HOT]
        if ObjectStoreClass(storage_class) in map:
            return map[ObjectStoreClass(storage_class)]
        return storage_class

    def convert_properties_metadata(
        self, properties: dict | None, metadata: dict | None
    ) -> dict:
        args: dict = {}
        if properties is None:
            props = ObjectProperties()
        else:
            props = ObjectProperties(**properties)
        if props.cache_control is not None:
            args["CacheControl"] = props.cache_control
        if props.content_disposition is not None:
            args["ContentDisposition"] = props.content_disposition
        if props.content_encoding is not None:
            args["ContentEncoding"] = props.content_encoding
        if props.content_language is not None:
            args["ContentLanguage"] = props.content_language
        if props.content_length is not None:
            args["ContentLength"] = props.content_length
        if props.content_md5 is not None:
            args["ContentMD5"] = props.content_md5
        if props.content_type is not None:
            args["ContentType"] = props.content_type
        if props.crc32c is not None:
            args["ChecksumCRC32C"] = props.crc32c
        if props.expires is not None:
            args["Expires"] = props.expires
        if props.storage_class is not None:
            args["StorageClass"] = self.convert_storage_class(
                props.storage_class
            )
        if metadata is not None:
            args["Metadata"] = metadata
        return args

    def convert_put(
        self,
        id: str,
        value: bytes | None,
        file: str | None,
        stream: IO | None,
        metadata: dict | None,
        properties: dict | None,
        match_condition: MatchCondition,
        config: ObjectTransferConfig | None,
    ) -> tuple[dict, str, dict]:
        args: dict = {}
        state: dict = {}
        func = "put_object"
        args["Bucket"] = self.bucket
        args["Key"] = id
        if config is not None and config.multipart:
            if value is not None:
                func = "upload_fileobj"
                args["Fileobj"] = BytesIO(value)
            elif stream is not None:
                func = "upload_fileobj"
                args["Fileobj"] = stream
            elif file is not None:
                func = "upload_file"
                args["Filename"] = file
            args["ExtraArgs"] = self.convert_properties_metadata(
                properties, metadata
            )
            args["Config"] = self.convert_transfer_config(config)
        else:
            func = "put_object"
            if value is not None:
                args["Body"] = value
            elif stream is not None:
                args["Body"] = stream
            elif file is not None:
                args["Body"] = open(file, "rb")
                state["open_file"] = args["Body"]
            args = args | self.convert_properties_metadata(
                properties, metadata
            )
        return args, func, state

    def convert_get(
        self,
        id: str,
        version: str | None,
        file: str | None,
        stream: IO | None,
        match_condition: MatchCondition,
        start: int | None,
        end: int | None,
        config: ObjectTransferConfig | None,
    ) -> tuple[dict, str, dict]:
        args: dict = {}
        state: dict = {}
        func = "get_object"
        args["Bucket"] = self.bucket
        args["Key"] = id
        if config is not None and config.multipart:
            if file is not None:
                func = "download_file"
                args["Filename"] = file
            elif stream is not None:
                func = "download_fileobj"
                args["Fileobj"] = stream
            else:
                func = "download_fileobj"
                fileobj = BytesIO()
                state["fileobj"] = fileobj
                args["Fileobj"] = fileobj
            if version is not None:
                args["ExtraArgs"] = {"VersionId": version}
            args["Config"] = self.convert_transfer_config(config)
        else:
            func = "get_object"
            if version is not None:
                args["VersionId"] = version
            args = args | self.convert_match_condition(match_condition)
            if start is not None and end is not None:
                args["Range"] = f"bytes={start}-{end}"
            elif start is not None:
                args["Range"] = f"bytes={start}-"
            elif end is not None:
                args["Range"] = f"bytes=0-{end}"
        state["func"] = func
        return args, func, state

    def convert_match_condition(self, match_condition: MatchCondition) -> dict:
        args: dict = {}
        if match_condition.if_match is not None:
            args["IfMatch"] = match_condition.if_match
        if match_condition.if_none_match is not None:
            args["IfNoneMatch"] = match_condition.if_none_match
        if match_condition.if_modified_since is not None:
            args["IfModifiedSince"] = match_condition.if_modified_since
        if match_condition.if_unmodified_since is not None:
            args["IfUnmodifiedSince"] = match_condition.if_unmodified_since
        return args

    def convert_copy_match_condition(
        self, match_condition: MatchCondition
    ) -> dict:
        args: dict = {}
        if match_condition.if_match is not None:
            args["CopySourceIfMatch"] = match_condition.if_match
        if match_condition.if_none_match is not None:
            args["CopySourceIfNoneMatch"] = match_condition.if_none_match
        if match_condition.if_modified_since is not None:
            args["CopySourceIfModifiedSince"] = (
                match_condition.if_modified_since
            )
        if match_condition.if_unmodified_since is not None:
            args["CopySourceIfUnmodifiedSince"] = (
                match_condition.if_unmodified_since
            )
        return args

    def convert_get_properties(
        self,
        id: str,
        version: str | None,
        match_condition: MatchCondition,
    ) -> dict:
        args: dict = {"Bucket": self.bucket, "Key": id}
        if version is not None:
            args["VersionId"] = version
        args = args | self.convert_match_condition(match_condition)
        return args

    def convert_get_versions(self, id: str) -> dict:
        return {"Bucket": self.bucket, "Prefix": id}

    def convert_update(
        self,
        id: str,
        version: str | None,
        metadata: dict | None,
        properties: dict | None,
        match_condition: MatchCondition,
    ) -> dict:
        args: dict = {}
        args["Bucket"] = self.bucket
        args["Key"] = id
        args["CopySource"] = {"Bucket": self.bucket, "Key": id}
        if version is not None:
            args["CopySource"]["VersionId"] = version
        if metadata is not None:
            args["Metadata"] = metadata
            args["MetadataDirective"] = "REPLACE"
        args = args | self.convert_properties_metadata(properties, None)
        args = args | self.convert_copy_match_condition(match_condition)
        return args

    def convert_delete(self, id: str, version: str | None) -> tuple[dict, str]:
        if version == "*":
            args = {"id": id, "version": version}
            return args, "helper_delete"
        args = {"Bucket": self.bucket, "Key": id}
        if version is not None:
            args["VersionId"] = version
        return args, "delete_object"

    def convert_copy(
        self,
        id: str,
        source_id: str,
        source_version: str | None,
        source_collection: str | None,
        metadata: dict | None,
        properties: dict | None,
        match_condition: MatchCondition,
    ):
        args: dict = {}
        args["Bucket"] = self.bucket
        args["Key"] = id
        args["CopySource"] = {
            "Bucket": (
                source_collection
                if source_collection is not None
                else self.bucket
            ),
            "Key": source_id,
        }
        if source_version is not None:
            args["CopySource"]["VersionId"] = source_version
        if metadata is not None:
            args["Metadata"] = metadata
            args["MetadataDirective"] = "REPLACE"
        args = args | self.convert_properties_metadata(properties, None)
        args = args | self.convert_copy_match_condition(match_condition)
        return args

    def convert_generate(
        self,
        id: str,
        version: str | None,
        method: str | None,
        expiry: int | None,
    ):
        args: dict = {}
        args["Params"] = {"Bucket": self.bucket, "Key": id}
        if version is not None:
            args["Params"]["VersionId"] = version
        if method is None or method == "GET":
            args["ClientMethod"] = "get_object"
        elif method == "PUT":
            args["ClientMethod"] = "put_object"
        else:
            raise BadRequestError("Method not supported for generating url")
        if expiry is not None:
            args["ExpiresIn"] = expiry
        return args

    def convert_batch(
        self, op_parsers: list[StoreOperationParser]
    ) -> tuple[dict, str]:
        for op_parser in op_parsers:
            if not op_parser.op_equals(StoreOperation.DELETE):
                raise BadRequestError("Operation not supported in BATCH")
        return self.convert_batch_delete(op_parsers), "delete"

    def convert_batch_delete(
        self, op_parsers: list[StoreOperationParser]
    ) -> dict:
        args: dict = {}
        args["Bucket"] = self.bucket
        args["Delete"] = {"Objects": []}
        for op_parser in op_parsers:
            id = op_parser.get_id_as_str()
            version = op_parser.get_version()
            object = {"Key": id}
            if version is not None:
                object["VersionId"] = version
            args["Delete"]["Objects"].append(object)
        return args


class AmazonS3Collection:
    client: Any
    helper: Any
    op_converter: OperationConverter
    result_converter: ResultConverter
    processor: ItemProcessor

    def __init__(
        self,
        resource: Any,
        client: Any,
        bucket: str,
    ):
        self.client = client
        self.helper = ClientHelper(resource, client, bucket)
        self.processor = ItemProcessor()
        self.op_converter = OperationConverter(self.processor, bucket)
        self.result_converter = ResultConverter(self.processor, bucket)



from __future__ import annotations

import json
from datetime import datetime
from typing import IO, Any

from verse.core import Operation, OperationParser
from verse.core.exceptions import BadRequestError
from verse.ql import (
    And,
    Comparison,
    ComparisonOp,
    Expression,
    Field,
    Function,
    FunctionNamespace,
    Not,
    Or,
    OrderBy,
    Parameter,
    QLParser,
    Select,
    Update,
    UpdateOp,
    UpdateOperation,
    Value,
)

from ._attributes import (
    Attribute,
    KeyAttribute,
    SpecialAttribute,
    UpdateAttribute,
)
from ._functions import StoreFunctionName
from ._models import (
    ArrayIndex,
    AscIndex,
    BaseIndex,
    CompositeIndex,
    DescIndex,
    FieldIndex,
    GeospatialIndex,
    HashIndex,
    Index,
    MatchCondition,
    RangeIndex,
    SparseVectorIndex,
    TextIndex,
    TTLIndex,
    VectorIndex,
    WildcardIndex,
)
from ._operation import StoreOperation


class StoreOperationParser(OperationParser):
    ARG_STATEMENT = "statement"
    ARG_SELECT = "select"
    ARG_COLLECTION = "collection"
    ARG_SET = "set"
    ARG_WHERE = "where"
    ARG_ORDER_BY = "order_by"
    ARG_SEARCH = "search"

    _ql_cache: dict[Any, Any]
    operation_parsers: list[StoreOperationParser] | None

    def __init__(self, operation: Operation | None):
        super().__init__(operation)
        self._ql_cache = dict()
        self.operation_parsers = None
        if self.op_equals(StoreOperation.EXECUTE):
            super().__init__(self.parse_statement())

    def is_single_op(self) -> bool:
        operations = self.get_arg("operations")
        return operations is None

    def is_resource_op(self) -> bool:
        if self.get_name() in [
            StoreOperation.CREATE_COLLECTION,
            StoreOperation.DROP_COLLECTION,
            StoreOperation.LIST_COLLECTIONS,
            StoreOperation.CLOSE,
        ]:
            return True
        return False

    def is_collection_op(self) -> bool:
        if self.get_name() in [
            StoreOperation.GET,
            StoreOperation.PUT,
            StoreOperation.UPDATE,
            StoreOperation.DELETE,
            StoreOperation.QUERY,
            StoreOperation.COUNT,
            StoreOperation.BATCH,
            StoreOperation.TRANSACT,
            StoreOperation.WATCH,
            StoreOperation.CREATE_INDEX,
            StoreOperation.DROP_INDEX,
            StoreOperation.LIST_INDEXES,
        ]:
            return True
        return False

    def parse_statement(self) -> Operation:
        operation = self.get_parsed_arg(StoreOperationParser.ARG_STATEMENT)
        if operation is None:
            raise BadRequestError("No operation found")
        params = self.get_params()
        operation = self.replace_operation_params(operation, params)
        if params is not None:
            operation.args["params"] = params
        return operation

    def get_params(self) -> dict[str, Any] | None:
        return self.get_arg("params")

    def get_operations(self) -> list[Operation]:
        return self.get_arg("operations")

    def get_operation_parsers(self) -> list[StoreOperationParser]:
        if self.operation_parsers is None:
            self.operation_parsers = []
            for operation in self.get_operations():
                params = self.get_params()
                if params is not None:
                    args: dict | None = {"params": params}
                    if operation.args is not None and args is not None:
                        args = operation.args | args
                else:
                    args = operation.args
                self.operation_parsers.append(
                    StoreOperationParser(
                        self.replace_operation_params(
                            Operation(name=operation.name, args=args), params
                        )
                    )
                )
        return self.operation_parsers

    def get_index(self) -> Index:
        index = self.get_arg("index")
        if index is None:
            raise BadRequestError("Index parameter missing")
        if isinstance(index, BaseIndex):
            return index
        type_map: dict = {
            "field": FieldIndex,
            "range": RangeIndex,
            "hash": HashIndex,
            "composite": CompositeIndex,
            "asc": AscIndex,
            "desc": DescIndex,
            "array": ArrayIndex,
            "ttl": TTLIndex,
            "text": TextIndex,
            "geospatial": GeospatialIndex,
            "vector": VectorIndex,
            "sparse_vector": SparseVectorIndex,
            "wildcard": WildcardIndex,
        }
        if isinstance(index, dict):
            if "type" in index and index["type"] in type_map:
                type = type_map[index["type"]]
                return type.from_dict(index)
        raise BadRequestError("Index format error")

    def get_key(self) -> Value:
        return self.get_arg("key", True)

    def get_id_from_key(self, key: Value) -> Value:
        if key is not None:
            if isinstance(key, (str, int, float, bool)):
                return key
            if isinstance(key, dict) and KeyAttribute.ID in key:
                return key[KeyAttribute.ID]
        return None

    def get_id(self) -> Value:
        return self.get_id_from_key(self.get_key())

    def get_id_as_str(self) -> str:
        id = self.get_id()
        if not isinstance(id, str):
            raise BadRequestError("id is not string")
        return id

    def get_expiry(self) -> int | None:
        return self.get_arg("expiry")

    def get_method(self) -> str | None:
        return self.get_arg("method")

    def get_id_as_str_or_none(self) -> str | None:
        id = self.get_id()
        if id is None:
            return None
        if not isinstance(id, str):
            raise BadRequestError("id is not string")
        return id

    def get_version_from_key(self, key: Value) -> str | None:
        if key is not None:
            if isinstance(key, dict) and KeyAttribute.VERSION in key:
                return key[KeyAttribute.VERSION]
        return None

    def get_version(self) -> str | None:
        return self.get_version_from_key(self.get_key())

    def get_label(self) -> str | None:
        key = self.get_key()
        if key is not None:
            if isinstance(key, dict) and KeyAttribute.LABEL in key:
                return key[KeyAttribute.LABEL]
        return None

    def get_source(self) -> Value:
        return self.get_arg("source", True)

    def get_source_collection(self) -> str | None:
        source = self.get_source()
        if isinstance(source, dict):
            if "collection" in source:
                return source["collection"]
        return None

    def get_source_key(self) -> Value:
        source = self.get_source()
        if isinstance(source, dict):
            if Attribute.KEY in source:
                return source[Attribute.KEY]
        return source

    def get_source_id_as_str(self) -> str:
        source = self.get_source()
        if isinstance(source, str):
            return source
        elif isinstance(source, dict) and Attribute.KEY in source:
            id = self.get_id_from_key(source[Attribute.KEY])
            if isinstance(id, str):
                return id
        raise BadRequestError("id is not string")

    def get_source_version(self) -> str | None:
        return self.get_version_from_key(self.get_source_key())

    def get_attribute(self) -> str:
        attr = self.get_arg("attr")
        if attr is not None:
            return attr
        return Attribute.VALUE

    def is_attribute(self, attribute: str) -> bool:
        return attribute.lower() == self.get_attribute().lower()

    def is_value_attribute(self) -> bool:
        return self.is_attribute(Attribute.VALUE)

    def is_metadata_attribute(self) -> bool:
        return self.is_attribute(Attribute.METADATA)

    def is_properties_attribute(self) -> bool:
        return self.is_attribute(Attribute.PROPERTIES)

    def is_versions_attribute(self) -> bool:
        return self.is_attribute(Attribute.VERSIONS)

    def get_filter(self) -> dict | None:
        return self.get_arg("filter")

    def get_select(self) -> Select | None:
        return self.get_parsed_arg(StoreOperationParser.ARG_SELECT)

    def get_collection_name(self) -> str | None:
        collection = self.get_arg("collection")
        if isinstance(collection, str):
            return collection
        collection = self.get_parsed_arg(StoreOperationParser.ARG_COLLECTION)
        if collection is None:
            return None
        return collection.name

    def get_value(self) -> Any:
        return self.get_arg("value", True)

    def get_value_as_bytes(self) -> bytes | None:
        value = self.get_value()
        if value is None:
            return None
        if isinstance(value, bytes):
            return value
        return json.dumps(value).encode()

    def get_file(self) -> str | None:
        return self.get_arg("file")

    def get_url(self) -> str | None:
        return self.get_arg("url")

    def get_stream(self) -> IO | None:
        return self.get_arg("stream")

    def get_start(self) -> int | None:
        return self.get_arg("start")

    def get_end(self) -> int | None:
        return self.get_arg("end")

    def get_vector_value(self) -> dict:
        value = self.get_value()
        if isinstance(value, list):
            return {"vector": value}
        if isinstance(value, dict):
            return value
        raise BadRequestError("Vector value must be list or dict")

    def get_metadata(self) -> dict | None:
        return self.get_arg("metadata")

    def get_properties(self) -> dict | None:
        return self.get_arg("properties")

    def get_set(self) -> Update:
        set = self.get_parsed_arg(StoreOperationParser.ARG_SET)
        return set

    def get_update_attribute(self) -> str:
        set = self.get_parsed_arg(StoreOperationParser.ARG_SET)
        if len(set.operations) == 1:
            return set.operations[0].field
        raise BadRequestError("Invalid number of update attributes")

    def is_value_update(self) -> bool:
        attribute = self.get_update_attribute()
        return attribute.lower() == UpdateAttribute.VALUE

    def is_metadata_update(self) -> bool:
        attribute = self.get_update_attribute()
        return attribute.lower() == UpdateAttribute.METADATA

    def is_properties_update(self) -> bool:
        attribute = self.get_update_attribute()
        return attribute.lower() == UpdateAttribute.PROPERTIES

    def get_updated_value(self) -> Any:
        set = self.get_parsed_arg(StoreOperationParser.ARG_SET)
        for operation in set.operations:
            if operation.field == UpdateAttribute.VALUE and (
                operation.op == UpdateOp.INSERT or operation.op == UpdateOp.PUT
            ):
                return operation.args[0]
        raise BadRequestError("Update value not found")

    def get_updated_metadata(self) -> dict | None:
        set = self.get_parsed_arg(StoreOperationParser.ARG_SET)
        for operation in set.operations:
            if operation.field == UpdateAttribute.METADATA and (
                operation.op == UpdateOp.INSERT or operation.op == UpdateOp.PUT
            ):
                return operation.args[0]
        raise BadRequestError("Update metadata not found")

    def get_updated_properities(self) -> dict | None:
        set = self.get_parsed_arg(StoreOperationParser.ARG_SET)
        for operation in set.operations:
            if operation.field == UpdateAttribute.PROPERTIES and (
                operation.op == UpdateOp.INSERT or operation.op == UpdateOp.PUT
            ):
                return operation.args[0]
        raise BadRequestError("Update properties not found")

    def get_search_as_function(self) -> Function | None:
        funcs = self.get_search_as_functions()
        if funcs is None:
            return None
        if funcs is not None and len(funcs) > 1:
            raise BadRequestError("Multiple search functions found.")
        return funcs[0]

    def get_search_as_functions(self) -> list[Function] | None:
        def _get_functions(expr):
            if expr is None:
                return None
            funcs = []
            if isinstance(expr, Function):
                return [expr]
            elif isinstance(expr, And):
                funcs.extend(_get_functions(expr.lexpr))
                funcs.extend(_get_functions(expr.rexpr))
            else:
                raise BadRequestError(
                    f"Expression not support in search {expr}"
                )
            return funcs

        expr = self.get_search()
        return _get_functions(expr)

    def get_search(self) -> Expression | None:
        expr = self.get_parsed_arg(StoreOperationParser.ARG_SEARCH)
        return expr

    def get_where(self) -> Expression | None:
        expr = self.get_parsed_arg(StoreOperationParser.ARG_WHERE)
        return expr

    def get_where_expr_list(self) -> list[Expression]:
        conditions = []

        def parse_expr(expr):
            if expr is None:
                return
            if isinstance(expr, And):
                parse_expr(expr.lexpr)
                parse_expr(expr.rexpr)
            elif isinstance(expr, Or):
                raise BadRequestError("OR is not supported in WHERE")
            elif isinstance(expr, Not):
                raise BadRequestError("NOT is not supported in WHERE")
            else:
                conditions.append(expr)

        expr = self.get_parsed_arg(StoreOperationParser.ARG_WHERE)
        parse_expr(expr)
        return conditions

    def get_match_condition(self) -> MatchCondition:
        match_condition = MatchCondition()
        expr_list = self.get_where_expr_list()
        for expr in expr_list:
            field = None
            if isinstance(expr, Function):
                if (
                    expr.namespace == FunctionNamespace.BUILTIN
                    and expr.name == StoreFunctionName.EXISTS
                ):
                    match_condition.exists = True
                if (
                    expr.namespace == FunctionNamespace.BUILTIN
                    and expr.name == StoreFunctionName.NOT_EXISTS
                ):
                    match_condition.exists = False
            if isinstance(expr, Comparison):
                if isinstance(expr.lexpr, Field):
                    field = expr.lexpr.path
                    value = expr.rexpr
                elif isinstance(expr.rexpr, Field):
                    field = expr.rexpr.path
                    value = expr.lexpr
                if field == SpecialAttribute.ETAG and isinstance(value, str):
                    if expr.op == ComparisonOp.EQ:
                        match_condition.if_match = value
                    elif expr.op == ComparisonOp.NEQ:
                        match_condition.if_none_match = value
                if field == SpecialAttribute.VERSION and isinstance(
                    value, str
                ):
                    if expr.op == ComparisonOp.EQ:
                        match_condition.if_version_match = value
                    elif expr.op == ComparisonOp.NEQ:
                        match_condition.if_version_not_match = value
                if field == SpecialAttribute.MODIFIED and isinstance(
                    value, datetime
                ):
                    if expr.op in [ComparisonOp.GT, ComparisonOp.GTE]:
                        match_condition.if_modified_since = value
                    elif expr.op in [ComparisonOp.LT, ComparisonOp.LTE]:
                        match_condition.if_unmodified_since = value
        return match_condition

    def get_where_exists(self) -> bool | None:
        expr = self.get_parsed_arg(StoreOperationParser.ARG_WHERE)
        if expr is not None:
            exists = True
            if isinstance(expr, Not):
                expr = expr.expr
                exists = False
            if isinstance(expr, Function):
                if (
                    expr.namespace == FunctionNamespace.BUILTIN
                    and expr.name == StoreFunctionName.EXISTS
                ):
                    return exists
                if (
                    expr.namespace == FunctionNamespace.BUILTIN
                    and expr.name == StoreFunctionName.NOT_EXISTS
                ):
                    return not exists
        return None

    def get_where_etag(self) -> Any:
        expr = self.get_parsed_arg(StoreOperationParser.ARG_WHERE)
        if (
            expr is not None
            and isinstance(expr, Comparison)
            and expr.op == ComparisonOp.EQ
        ):
            if (
                isinstance(expr.lexpr, Field)
                and expr.lexpr.path == SpecialAttribute.ETAG
            ):
                return expr.rexpr
            elif (
                isinstance(expr.rexpr, Field)
                and expr.rexpr.path == SpecialAttribute.ETAG
            ):
                return expr.lexpr
        return None

    def get_order_by(self) -> OrderBy | None:
        return self.get_parsed_arg(StoreOperationParser.ARG_ORDER_BY)

    def get_limit(self) -> int | None:
        return self.get_arg("limit")

    def get_offset(self) -> int | None:
        return self.get_arg("offset")

    def get_returning(self) -> bool | None:
        return self.get_arg("returning")

    def get_continuation(self) -> str | None:
        return self.get_arg("continuation")

    def get_config(self) -> Any:
        return self.get_arg("config")

    def get_nflags(self) -> dict | None:
        return self.get_arg("nflags")

    def get_parsed_arg(self, arg: Any) -> Any:
        if arg in self._ql_cache:
            return self._ql_cache[arg]
        if arg == StoreOperationParser.ARG_STATEMENT:
            statement = self.get_arg("statement")
            parsed_arg: Any = QLParser.parse_statement(statement)
        if arg == StoreOperationParser.ARG_SELECT:
            select = self.get_arg("select")
            if not isinstance(select, str):
                return select
            parsed_arg = QLParser.parse_select(select)
        elif arg == StoreOperationParser.ARG_COLLECTION:
            collection = self.get_arg("collection")
            if not isinstance(collection, str):
                return collection
            parsed_arg = QLParser.parse_collection(collection)
        elif arg == StoreOperationParser.ARG_SET:
            set = self.get_arg("set")
            if not isinstance(set, str):
                parsed_arg = set
            else:
                parsed_arg = QLParser.parse_update(set)
            parsed_arg = self.replace_set_params(parsed_arg, self.get_params())
        elif arg == StoreOperationParser.ARG_SEARCH:
            search = self.get_arg("search")
            if not isinstance(search, str):
                parsed_arg = search
            else:
                parsed_arg = QLParser.parse_search(search)
            parsed_arg = self.replace_expr_params(
                parsed_arg, self.get_params()
            )
        elif arg == StoreOperationParser.ARG_WHERE:
            where = self.get_arg("where")
            if not isinstance(where, str):
                parsed_arg = where
            else:
                parsed_arg = QLParser.parse_where(where)
            parsed_arg = self.replace_expr_params(
                parsed_arg, self.get_params()
            )
        elif arg == StoreOperationParser.ARG_ORDER_BY:
            order_by = self.get_arg("order_by")
            if not isinstance(order_by, str):
                return order_by
            parsed_arg = QLParser.parse_order_by(order_by)
        self._ql_cache[arg] = parsed_arg
        return parsed_arg

    def replace_operation_params(
        self, operation: Operation, params: dict[str, Any] | None
    ) -> Operation:
        if params is None or len(params) == 0:
            return operation
        if operation.args is not None:
            args: dict = {}
            for key, value in operation.args.items():
                args[key] = self.replace_expr_params(value, params)
            return Operation(name=operation.name, args=args)
        return Operation(name=operation.name)

    def replace_set_params(
        self, set: Update, params: dict[str, Any] | None
    ) -> Update:
        if params is None or len(params) == 0:
            return set
        copy = Update()
        for operation in set.operations:
            copy.operations.append(
                UpdateOperation(
                    field=operation.field,
                    op=operation.op,
                    args=tuple(
                        [
                            self.replace_expr_params(arg, params)
                            for arg in operation.args
                        ]
                    ),
                )
            )
        return copy

    def replace_expr_params(self, expr, params: dict[str, Any] | None) -> Any:
        if params is None or len(params) == 0:
            return expr
        if isinstance(expr, list):
            return [self.replace_expr_params(i, params) for i in expr]
        if isinstance(expr, Parameter):
            if expr.name in params:
                return params[expr.name]
            else:
                raise BadRequestError(f"Parameter {expr.name} not found")
        if isinstance(expr, Comparison):
            return Comparison(
                lexpr=self.replace_expr_params(expr.lexpr, params),
                op=expr.op,
                rexpr=self.replace_expr_params(expr.rexpr, params),
            )

        if isinstance(expr, Function):
            args = None
            named_args = None
            if expr.args is not None:
                args = tuple(
                    self.replace_expr_params(arg, params) for arg in expr.args
                )
            if expr.named_args is not None:
                named_args = dict()
                for k, v in expr.named_args.items():
                    named_args[k] = self.replace_expr_params(v, params)

            return Function(
                namespace=expr.namespace,
                name=expr.name,
                args=args,
                named_args=named_args,
            )
        if isinstance(expr, And):
            return And(
                lexpr=self.replace_expr_params(expr.lexpr, params),
                rexpr=self.replace_expr_params(expr.rexpr, params),
            )
        if isinstance(expr, Or):
            return Or(
                lexpr=self.replace_expr_params(expr.lexpr, params),
                rexpr=self.replace_expr_params(expr.rexpr, params),
            )
        if isinstance(expr, Not):
            return Not(expr=self.replace_expr_params(expr.expr, params))
        return expr



"""
Object Store on Google Cloud Storage.
"""

from __future__ import annotations

__all__ = ["GoogleCloudStorage"]

from datetime import timedelta
from io import BytesIO
from typing import IO, Any

from google.api_core.exceptions import NotFound
from google.api_core.exceptions import NotModified as NativeNotModified
from google.api_core.exceptions import PreconditionFailed
from google.cloud.storage import Blob, Bucket, Client

from verse.core import Context, NCall, Operation, Response
from verse.core.exceptions import (
    BadRequestError,
    NotFoundError,
    NotModified,
    PreconditionFailedError,
)
from verse.internal.storage_core import (
    ItemProcessor,
    MatchCondition,
    StoreOperation,
    StoreOperationParser,
    StoreProvider,
)

from .._helper import (
    QueryArgs,
    get_collection_config,
    get_query_args,
    get_transfer_config,
)
from .._models import (
    ObjectCollectionConfig,
    ObjectItem,
    ObjectKey,
    ObjectList,
    ObjectProperties,
    ObjectStoreClass,
    ObjectTransferConfig,
    ObjectVersion,
)


class GoogleCloudStorage(StoreProvider):
    project: str | None
    service_account_info: str | None
    service_account_file: str | None
    access_token: str | None
    bucket: str | None
    nparams: dict[str, Any]

    _credentials: Any
    _storage_client: Any

    _collection_cache: dict[str, GoogleCloudStorageCollection]

    def __init__(
        self,
        project: str | None = None,
        service_account_info: str | None = None,
        service_account_file: str | None = None,
        access_token: str | None = None,
        bucket: str | None = None,
        nparams: dict[str, Any] = dict(),
        **kwargs,
    ):
        """Initialize.

        Args:
            project:
                Google Cloud project name.
            service_account_info:
                Google service account info with serialized credentials.
            service_account_file:
                Google service account file with credentials.
            access_token:
                Google access token.
            bucket:
                Google Cloud Storage bucket name
                mapped to object store collection.
            nparams:
                Native params to Google Cloud Storage client.
        """
        self.project = project
        self.service_account_info = service_account_info
        self.service_account_file = service_account_file
        self.access_token = access_token
        self.bucket = bucket
        self.nparams = nparams

        self._storage_client = None
        self._collection_cache = dict()

    def init(self, context: Context | None = None) -> None:
        if self._storage_client is not None:
            return

        (
            self._credentials,
            self._storage_client,
        ) = self._init_credentials_client()

    def _init_credentials_client(
        self,
    ) -> tuple[Any, Any]:
        from verse.internal.storage_common._google import get_credentials

        credentials = get_credentials(
            self.service_account_info,
            self.service_account_file,
            self.access_token,
        )

        args = dict()
        if self.project is not None:
            args["project"] = self.project
        if credentials is not None:
            args["credentials"] = credentials
        if self.nparams is not None:
            args = args | self.nparams
        client: Any = Client(**args)
        return credentials, client

    def _get_bucket_name(self, op_parser: StoreOperationParser) -> str | None:
        collection_name = (
            op_parser.get_operation_parsers()[0].get_collection_name()
            if op_parser.op_equals(StoreOperation.BATCH)
            else op_parser.get_collection_name()
        )
        bucket = (
            collection_name if collection_name is not None else self.bucket
        )
        return bucket

    def _get_collection(
        self, op_parser: StoreOperationParser
    ) -> GoogleCloudStorageCollection | None:
        if op_parser.is_resource_op():
            return None
        bucket_name = self._get_bucket_name(op_parser)
        if bucket_name is None:
            raise BadRequestError("Collection name must be specified")
        if bucket_name in self._collection_cache:
            return self._collection_cache[bucket_name]
        client = self._storage_client.bucket(bucket_name)
        col = GoogleCloudStorageCollection(
            self._storage_client, client, bucket_name
        )
        self._collection_cache[bucket_name] = col
        return col

    def _validate(self, op_parser: StoreOperationParser):
        pass

    def run(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
    ) -> Any:
        self.init(context=context)
        op_parser = self.get_op_parser(operation)
        self._validate(op_parser)
        collection = self._get_collection(op_parser)
        ncall, state = self._get_ncall(
            op_parser,
            collection,
            ResourceHelper(self._storage_client),
        )
        if ncall is None:
            return super().run(operation=operation, context=context)
        nresult = ncall.invoke()
        result = self._convert_nresult(nresult, state, op_parser, collection)
        return Response(result=result, native=dict(result=nresult, call=ncall))

    def _get_ncall(
        self,
        op_parser: StoreOperationParser,
        collection: GoogleCloudStorageCollection | None,
        resource_helper: Any,
    ) -> tuple[NCall | None, dict]:
        if collection is not None:
            op_converter = collection.op_converter
            client = collection.client
            helper = collection.helper
        call = None
        state: dict = {}
        nargs = op_parser.get_nargs()
        # CREATE COLLECTION
        if op_parser.op_equals(StoreOperation.CREATE_COLLECTION):
            args = {
                "bucket": self._get_bucket_name(op_parser),
                "config": get_collection_config(op_parser),
                "nargs": nargs,
            }
            call = NCall(resource_helper.create_collection, args)
        # DROP COLLECTION
        elif op_parser.op_equals(StoreOperation.DROP_COLLECTION):
            args = {
                "bucket": self._get_bucket_name(op_parser),
                "nargs": nargs,
            }
            call = NCall(resource_helper.drop_collection, args)
        # LIST COLLECTIONS
        elif op_parser.op_equals(StoreOperation.LIST_COLLECTIONS):
            args = {"nargs": nargs}
            call = NCall(resource_helper.list_collections, args)
        # PUT
        elif op_parser.op_equals(StoreOperation.PUT):
            args, func, blob = op_converter.convert_put(
                id=op_parser.get_id_as_str(),
                value=op_parser.get_value_as_bytes(),
                file=op_parser.get_file(),
                stream=op_parser.get_stream(),
                metadata=op_parser.get_metadata(),
                properties=op_parser.get_properties(),
                match_condition=op_parser.get_match_condition(),
                returning=op_parser.get_returning(),
                config=get_transfer_config(op_parser),
            )
            if func == "upload_from_file":
                call = NCall(
                    blob.upload_from_file,
                    args,
                    nargs,
                    {
                        NotFound: PreconditionFailedError,
                        PreconditionFailed: PreconditionFailedError,
                    },
                )
            elif func == "upload_from_filename":
                call = NCall(
                    blob.upload_from_filename,
                    args,
                    nargs,
                    {
                        NotFound: PreconditionFailedError,
                        PreconditionFailed: PreconditionFailedError,
                    },
                )
            elif func == "helper_put":
                args["nargs"] = nargs
                call = NCall(
                    helper.put,
                    args,
                    None,
                    {
                        NotFound: PreconditionFailedError,
                        PreconditionFailed: PreconditionFailedError,
                    },
                )
            state["blob"] = blob
        # GET value
        elif (
            op_parser.op_equals(StoreOperation.GET)
            and op_parser.is_value_attribute()
        ):
            args, func, blob = op_converter.convert_get(
                id=op_parser.get_id_as_str(),
                version=op_parser.get_version(),
                file=op_parser.get_file(),
                stream=op_parser.get_stream(),
                match_condition=op_parser.get_match_condition(),
                start=op_parser.get_start(),
                end=op_parser.get_end(),
                config=get_transfer_config(op_parser),
            )
            if func == "download_to_filename":
                call = NCall(
                    blob.download_to_filename,
                    args,
                    nargs,
                    {
                        NotFound: NotFoundError,
                        PreconditionFailed: PreconditionFailedError,
                        NativeNotModified: NotModified,
                    },
                )
            elif func == "download_to_file":
                call = NCall(
                    blob.download_to_file,
                    args,
                    nargs,
                    {
                        NotFound: NotFoundError,
                        PreconditionFailed: PreconditionFailedError,
                        NativeNotModified: NotModified,
                    },
                )
            elif func == "download_as_bytes":
                call = NCall(
                    blob.download_as_bytes,
                    args,
                    nargs,
                    {
                        NotFound: NotFoundError,
                        PreconditionFailed: PreconditionFailedError,
                        NativeNotModified: NotModified,
                    },
                )
            elif func == "helper_get":
                args["nargs"] = nargs
                call = NCall(
                    helper.get,
                    args,
                    None,
                    {
                        NotFound: NotFoundError,
                        PreconditionFailed: PreconditionFailedError,
                        NativeNotModified: NotModified,
                    },
                )
            state["blob"] = blob
            state["func"] = func
        # GET metadata or properties
        elif op_parser.op_equals(StoreOperation.GET) and (
            op_parser.is_metadata_attribute()
            or op_parser.is_properties_attribute()
        ):
            args = op_converter.convert_get_properties(
                id=op_parser.get_id_as_str(),
                version=op_parser.get_version(),
                match_condition=op_parser.get_match_condition(),
            )
            call = NCall(
                client.get_blob,
                args,
                nargs,
                {
                    NotFound: NotFoundError,
                    None: NotFoundError,
                    PreconditionFailed: PreconditionFailedError,
                    NativeNotModified: NotModified,
                },
            )
        # GET versions
        elif op_parser.op_equals(StoreOperation.GET) and (
            op_parser.is_versions_attribute()
        ):
            args = op_converter.convert_get_versions(
                id=op_parser.get_id_as_str()
            )
            call = NCall(
                client.list_blobs,
                args,
                nargs,
                {
                    NotFound: NotFoundError,
                    PreconditionFailed: PreconditionFailedError,
                },
            )
        # UPDATE
        elif op_parser.op_equals(StoreOperation.UPDATE):
            args, blob, update_storage_class, update_others = (
                op_converter.convert_update(
                    id=op_parser.get_id_as_str(),
                    version=op_parser.get_version(),
                    metadata=op_parser.get_updated_metadata(),
                    properties=op_parser.get_updated_properities(),
                    match_condition=op_parser.get_match_condition(),
                )
            )
            call = NCall(
                helper.update,
                {
                    "blob": blob,
                    "args": args,
                    "update_storage_class": update_storage_class,
                    "update_others": update_others,
                    "nargs": nargs,
                },
                nargs,
                {
                    NotFound: NotFoundError,
                    PreconditionFailed: PreconditionFailedError,
                },
            )
            state["blob"] = blob
        # DELETE
        elif op_parser.op_equals(StoreOperation.DELETE):
            args, func = op_converter.convert_delete(
                id=op_parser.get_id_as_str(),
                version=op_parser.get_version(),
                match_condition=op_parser.get_match_condition(),
            )
            if func == "helper_delete":
                args["nargs"] = nargs
                call = NCall(
                    helper.delete,
                    args,
                    None,
                    {
                        NotFound: None,
                        PreconditionFailed: PreconditionFailedError,
                    },
                )
            elif func == "delete_blob":
                call = NCall(
                    client.delete_blob,
                    args,
                    nargs,
                    {
                        NotFound: None,
                        PreconditionFailed: PreconditionFailedError,
                    },
                )
        # COPY
        elif op_parser.op_equals(StoreOperation.COPY):
            args, source_bucket = op_converter.convert_copy(
                id=op_parser.get_id_as_str(),
                source_id=op_parser.get_source_id_as_str(),
                source_version=op_parser.get_source_version(),
                source_collection=op_parser.get_source_collection(),
                metadata=op_parser.get_metadata(),
                properties=op_parser.get_properties(),
                match_condition=op_parser.get_match_condition(),
            )
            call = NCall(
                source_bucket.copy_blob,
                args,
                nargs,
                {
                    NotFound: NotFoundError,
                    PreconditionFailed: PreconditionFailedError,
                },
            )
        # GENERATE signed url
        elif op_parser.op_equals(StoreOperation.GENERATE):
            args, blob = op_converter.convert_generate(
                id=op_parser.get_id_as_str(),
                version=op_parser.get_version(),
                method=op_parser.get_method(),
                expiry=op_parser.get_expiry(),
            )
            call = NCall(blob.generate_signed_url, args, nargs)
        # QUERY
        elif op_parser.op_equals(StoreOperation.QUERY):
            query_args = get_query_args(op_parser)
            state["query_args"] = query_args
            args = {"query_args": get_query_args(op_parser), "nargs": nargs}
            call = NCall(helper.query, args, None)
        # COUNT
        elif op_parser.op_equals(StoreOperation.COUNT):
            query_args = get_query_args(op_parser)
            state["query_args"] = query_args
            args = {"query_args": get_query_args(op_parser), "nargs": nargs}
            call = NCall(helper.query, args, None)
        # BATCH
        elif op_parser.op_equals(StoreOperation.BATCH):
            args, func = op_converter.convert_batch(
                op_parser.get_operation_parsers()
            )
            if func == "delete":
                args["nargs"] = nargs
                call = NCall(helper.batch_delete, args, None)
            state["func"] = func
        # CLOSE
        elif op_parser.op_equals(StoreOperation.CLOSE):
            args = {"nargs": nargs}
            call = NCall(resource_helper.close, args)
        return call, state

    def _convert_nresult(
        self,
        nresult: Any,
        state: dict,
        op_parser: StoreOperationParser,
        collection: GoogleCloudStorageCollection | None,
    ) -> Any:
        if collection is not None:
            result_converter = collection.result_converter
        result: Any = None
        # CREATE COLLECTION
        if op_parser.op_equals(StoreOperation.CREATE_COLLECTION):
            result = None
        # DROP COLLECTION
        elif op_parser.op_equals(StoreOperation.DROP_COLLECTION):
            result = None
        # LIST COLLECTIONS
        elif op_parser.op_equals(StoreOperation.LIST_COLLECTIONS):
            result = [b.name for b in nresult]
        # PUT
        elif op_parser.op_equals(StoreOperation.PUT):
            blob = state["blob"]
            result = result_converter.convert_put_result(
                op_parser.get_id_as_str(), blob
            )
        # GET value
        elif (
            op_parser.op_equals(StoreOperation.GET)
            and op_parser.is_value_attribute()
        ):
            result = result_converter.convert_get_result(
                op_parser.get_id_as_str(),
                op_parser.get_file(),
                op_parser.get_stream(),
                nresult,
                state,
            )
        # GET metadata or properties
        elif op_parser.op_equals(StoreOperation.GET) and (
            op_parser.is_metadata_attribute()
            or op_parser.is_properties_attribute()
        ):
            result = result_converter.convert_get_properties_result(
                op_parser.get_id_as_str(), nresult
            )
        # GET versions
        elif op_parser.op_equals(StoreOperation.GET) and (
            op_parser.is_versions_attribute()
        ):
            result = result_converter.convert_get_versions_result(
                op_parser.get_id_as_str(), nresult
            )
        # UPDATE
        elif op_parser.op_equals(StoreOperation.UPDATE):
            blob = state["blob"]
            result = result_converter.convert_update_result(
                op_parser.get_id_as_str(), blob
            )
        # DELETE
        elif op_parser.op_equals(StoreOperation.DELETE):
            result = None
        # COPY
        elif op_parser.op_equals(StoreOperation.COPY):
            result = result_converter.convert_copy_result(
                op_parser.get_id_as_str(), nresult
            )
        # GENERATE signed url
        elif op_parser.op_equals(StoreOperation.GENERATE):
            result = result_converter.convert_generate_result(
                op_parser.get_id_as_str(),
                op_parser.get_version(),
                nresult,
            )
        # QUERY
        elif op_parser.op_equals(StoreOperation.QUERY):
            result = result_converter.convert_query_result(
                nresult,
                state["query_args"],
            )
        # COUNT
        elif op_parser.op_equals(StoreOperation.COUNT):
            result = result_converter.convert_count_result(
                nresult,
            )
        # BATCH
        elif op_parser.op_equals(StoreOperation.BATCH):
            result = []
            func = state["func"]
            if func == "delete":
                for op_parser in op_parser.get_operation_parsers():
                    result.append(None)
        return result


class ResourceHelper:
    storage_client: Any

    def __init__(self, storage_client: Any):
        self.storage_client = storage_client

    def create_collection(
        self, bucket: str, config: ObjectCollectionConfig | None, nargs: Any
    ):
        storage_bucket = self.storage_client.bucket(bucket)
        if config is not None and config.versioned is True:
            storage_bucket.versioning_enabled = True
        args = {"bucket_or_name": storage_bucket}
        if config is not None and config.nconfig is not None:
            args = args | config.nconfig
        storage_bucket = NCall(
            self.storage_client.create_bucket,
            args,
            nargs,
        ).invoke()
        if config is not None and config.acl is not None:
            if config.acl == "public-read":
                storage_bucket = NCall(
                    storage_bucket.make_public,
                    {"future": True},
                    None,
                ).invoke()

    def drop_collection(self, bucket: str, nargs: Any):
        gbucket = self.storage_client.bucket(bucket)
        args = {"force": True}
        NCall(
            gbucket.delete,
            args,
            nargs,
        ).invoke()

    def list_collections(self, nargs) -> Any:
        response = NCall(
            self.storage_client.list_buckets, None, nargs
        ).invoke()
        return response

    def close(self, nargs: Any) -> Any:
        pass


class ClientHelper:
    storage_client: Any
    client: Any
    bucket: str
    op_converter: OperationConverter

    def __init__(
        self,
        storage_client: Any,
        client: Any,
        bucket: str,
        op_converter: OperationConverter,
    ) -> None:
        self.storage_client = storage_client
        self.client = client
        self.bucket = bucket
        self.op_converter = op_converter

    def put(
        self,
        blob: Blob,
        value: bytes | None,
        file: str | None,
        stream: IO | None,
        match_condition: MatchCondition,
        returning: bool | None,
        config: ObjectTransferConfig | None,
        nargs: Any,
    ) -> Any:
        import os
        import tempfile

        from google.cloud.storage import transfer_manager

        raw_value = None
        filename = None
        args: dict = {}
        args["blob"] = blob
        if config is not None:
            if config.chunksize is not None:
                args["chunk_size"] = config.chunksize
            if config.concurrency is not None:
                args["max_workers"] = config.concurrency
        if file is not None:
            filename = file
        elif value is not None:
            raw_value = value
        elif stream is not None:
            raw_value = stream.read()
        if raw_value is not None:
            with tempfile.NamedTemporaryFile(delete=False) as temp_file:
                temp_file.write(raw_value)
                filename = temp_file.name
        if filename is None:
            raise BadRequestError("Object not provided")
        args["filename"] = filename
        NCall(
            transfer_manager.upload_chunks_concurrently, args, nargs
        ).invoke()
        if raw_value is not None:
            os.remove(filename)
        if returning is True:
            NCall(blob.reload, None, None).invoke()

    def get(
        self,
        blob: Blob,
        file: str | None,
        stream: IO | None,
        match_condition: MatchCondition,
        config: ObjectTransferConfig | None,
        nargs: Any,
    ) -> Any:
        import os
        import shutil
        import tempfile

        from google.cloud.storage import transfer_manager

        generated = False
        folder = None
        filename = None
        args: dict = {}
        args["blob"] = blob
        if config is not None:
            if config.chunksize is not None:
                args["chunk_size"] = config.chunksize
            if config.concurrency is not None:
                args["max_workers"] = config.concurrency
        if file is not None:
            filename = file
        else:
            generated = True
            folder = tempfile.mkdtemp()
            filename = os.path.join(folder, "temp")
        args["filename"] = filename
        args["download_kwargs"] = self.op_converter.convert_match_condition(
            match_condition
        )
        NCall(
            transfer_manager.download_chunks_concurrently, args, nargs
        ).invoke()
        if generated:
            with open(filename, "rb") as f:
                if stream is not None:
                    stream.write(f.read())
                else:
                    return f.read()
            if folder is not None:
                shutil.rmtree(folder)

    def update(
        self,
        blob: Blob,
        args: dict,
        update_storage_class: bool,
        update_others: bool,
        nargs: Any,
    ) -> Any:
        result = None
        if update_storage_class:
            new_class = blob.storage_class
            blob.storage_class = None
            args["new_class"] = new_class
            result = NCall(blob.update_storage_class, args, nargs).invoke()
            args.pop("new_class")
        if update_others:
            result = NCall(blob.update, args, nargs).invoke()
        return result

    def delete(
        self,
        id: str,
        version: str | None,
        match_condition: MatchCondition,
        nargs: Any,
    ) -> Any:
        if version == "*":
            blobs = self.client.list_blobs(prefix=id, versions=True)
            blobs_to_delete = []
            for blob in blobs:
                if blob.name == id:
                    blobs_to_delete.append(blob)
            if len(blobs_to_delete) > 1:
                with self.storage_client.batch():
                    for blob in blobs_to_delete:
                        if blob.name == id:
                            blob.delete()
        return None

    def query(
        self,
        query_args: QueryArgs,
        nargs: Any,
    ) -> Any:
        blobs: list = []
        prefixes: list = []
        args: dict = {}
        args["bucket_or_name"] = self.bucket
        if query_args.prefix is not None:
            args["prefix"] = query_args.prefix
        if query_args.delimiter is not None:
            args["delimiter"] = query_args.delimiter
            args["include_folders_as_prefixes"] = True
        if query_args.start_after is not None:
            args["start_offset"] = f"{query_args.start_after}~"
        if query_args.end_before is not None:
            args["end_offset"] = query_args.end_before
        if query_args.continuation is not None:
            args["page_token"] = query_args.continuation
        if query_args.paging:
            if (
                query_args.limit is not None
                and query_args.page_size is not None
            ):
                if query_args.page_size < query_args.limit:
                    args["max_results"] = query_args.page_size
                else:
                    args["max_results"] = query_args.limit
            elif query_args.limit is not None:
                args["max_results"] = query_args.limit
            elif query_args.page_size is not None:
                args["max_results"] = query_args.page_size
        else:
            if (
                query_args.limit is not None
                and query_args.page_size is not None
            ):
                args["max_results"] = query_args.limit
                args["page_size"] = query_args.page_size
            elif query_args.limit is not None:
                args["max_results"] = query_args.limit
            elif query_args.page_size is not None:
                args["page_size"] = query_args.page_size

        response = NCall(self.storage_client.list_blobs, args, nargs).invoke()
        for blob in response:
            blobs.append(blob)

        for prefix in response.prefixes:
            prefixes.append(prefix)

        prefixes = sorted(prefixes)
        continuation = None
        if query_args.page_size:
            continuation = response.next_page_token

        return dict(blobs=blobs, prefixes=prefixes, continuation=continuation)

    def batch_delete(self, objects: list[dict], nargs: Any) -> Any:
        with self.storage_client.batch():
            for object in objects:
                blob = self.client.blob(object["id"])
                if object["version"] is not None:
                    blob.generation = int(object["version"])
                blob.delete()


class ResultConverter:
    processor: ItemProcessor
    bucket: str

    def __init__(self, processor: ItemProcessor, bucket: str):
        self.processor = processor
        self.bucket = bucket

    def convert_storage_class(self, storage_class: str | None) -> str:
        map = {
            "STANDARD": ObjectStoreClass.HOT,
            "NEARLINE": ObjectStoreClass.COOL,
            "COLDLINE": ObjectStoreClass.COLD,
            "ARCHIVE": ObjectStoreClass.ARCHIVE,
        }
        if storage_class is None:
            return ObjectStoreClass.HOT.value
        if storage_class in map:
            return map[storage_class].value
        return storage_class

    def convert_properties(self, blob: Blob) -> ObjectProperties:
        properties = ObjectProperties()
        properties.cache_control = blob.cache_control
        properties.content_disposition = blob.content_disposition
        properties.content_encoding = blob.content_encoding
        properties.content_language = blob.content_language
        properties.content_length = blob.size
        properties.content_md5 = blob.md5_hash
        properties.content_type = blob.content_type
        properties.crc32c = blob.crc32c
        properties.last_modified = blob.updated
        properties.etag = blob.etag
        properties.storage_class = self.convert_storage_class(
            blob.storage_class
        )
        return properties

    def convert_put_result(self, id: str, blob: Blob) -> ObjectItem:
        return ObjectItem(
            key=ObjectKey(id=id, version=str(blob.generation)),
            metadata=blob.metadata,
            properties=self.convert_properties(blob),
            url=blob.public_url,
        )

    def convert_get_result(
        self,
        id: str,
        file: str | None,
        stream: IO | None,
        nresult: Any,
        state: dict,
    ) -> ObjectItem:
        value = None
        blob = state["blob"]

        if file is None and stream is None:
            value = nresult
        return ObjectItem(
            key=ObjectKey(id=id, version=str(blob.generation)),
            value=value,
            metadata=blob.metadata,
            properties=self.convert_properties(blob),
            url=blob.public_url,
        )

    def convert_get_properties_result(self, id: str, blob: Blob) -> ObjectItem:
        key = ObjectKey(id=id, version=str(blob.generation))
        properties = self.convert_properties(blob)
        return ObjectItem(
            key=key,
            metadata=blob.metadata,
            properties=properties,
            url=blob.public_url,
        )

    def convert_get_versions_result(self, id: str, nresult: Any) -> ObjectItem:
        key = ObjectKey(id=id)
        versions: list = []
        url = None
        for blob in nresult:
            if blob.name != id:
                continue
            properties = self.convert_properties(blob)
            versions.append(
                ObjectVersion(
                    version=str(blob.generation),
                    properties=properties,
                    metadata=blob.metadata,
                )
            )
            url = blob.public_url
        if len(versions) == 0:
            raise NotFoundError
        return ObjectItem(key=key, versions=versions, url=url)

    def convert_update_result(self, id: str, blob: Blob) -> ObjectItem:
        return ObjectItem(
            key=ObjectKey(id=id, version=str(blob.generation)),
            metadata=blob.metadata,
            properties=self.convert_properties(blob),
            url=blob.public_url,
        )

    def convert_copy_result(self, id: str, blob: Blob) -> ObjectItem:
        return ObjectItem(
            key=ObjectKey(id=id, version=str(blob.generation)),
            metadata=blob.metadata,
            properties=self.convert_properties(blob),
            url=blob.public_url,
        )

    def convert_generate_result(
        self, id: str, version: str | None, nresult: Any
    ) -> ObjectItem:
        return ObjectItem(
            key=ObjectKey(id=id, version=version),
            url=nresult,
        )

    def convert_query_result(
        self, nresult: Any, query_args: QueryArgs
    ) -> ObjectList:
        blobs = nresult["blobs"]
        prefixes = nresult["prefixes"]
        continuation = nresult["continuation"]
        items: list = []
        for blob in blobs:
            items.append(
                ObjectItem(
                    key=ObjectKey(id=blob.name),
                    properties=self.convert_properties(blob),
                    metadata=blob.metadata,
                    url=blob.public_url,
                )
            )
        return ObjectList(
            items=items, continuation=continuation, prefixes=prefixes
        )

    def convert_count_result(
        self,
        nresult: Any,
    ) -> int:
        blobs = nresult["blobs"]
        prefixes = nresult["prefixes"]
        return len(blobs) + len(prefixes)


class OperationConverter:
    storage_client: Any
    client: Any
    processor: ItemProcessor
    bucket: str

    def __init__(
        self,
        storage_client: Any,
        client: Any,
        processor: ItemProcessor,
        bucket: str,
    ):
        self.storage_client = storage_client
        self.client = client
        self.processor = processor
        self.bucket = bucket

    def convert_storage_class(self, storage_class: str | None) -> str:
        map = {
            ObjectStoreClass.HOT: "STANDARD",
            ObjectStoreClass.COOL: "NEARLINE",
            ObjectStoreClass.COLD: "COLDLINE",
            ObjectStoreClass.ARCHIVE: "ARCHIVE",
        }
        if storage_class is None:
            return map[ObjectStoreClass.HOT]
        if ObjectStoreClass(storage_class) in map:
            return map[ObjectStoreClass(storage_class)]
        return storage_class

    def convert_properties_metadata(
        self, blob: Blob, properties: dict | None, metadata: dict | None
    ) -> Blob:
        if properties is None:
            props = ObjectProperties()
        else:
            props = ObjectProperties(**properties)
        if props.cache_control is not None:
            blob.cache_control = props.cache_control
        if props.content_disposition is not None:
            blob.content_disposition = props.content_disposition
        if props.content_encoding is not None:
            blob.content_encoding = props.content_encoding
        if props.content_language is not None:
            blob.content_language = props.content_language
        if props.content_md5 is not None:
            blob.md5_hash = props.content_md5
        if props.content_type is not None:
            blob.content_type = props.content_type
        if props.crc32c is not None:
            blob.crc32c = props.crc32c
        if props.storage_class is not None:
            blob.storage_class = self.convert_storage_class(
                props.storage_class
            )
        if metadata is not None:
            blob.metadata = metadata
        return blob

    def convert_match_condition(self, match_condition: MatchCondition) -> dict:
        args: dict = {}
        if match_condition.exists is True:
            args["if_generation_match"] = 0
        if match_condition.exists is False:
            args["if_generation_not_match"] = 0
        if match_condition.if_match is not None:
            args["if_etag_match"] = match_condition.if_match
        if match_condition.if_none_match is not None:
            args["if_etag_not_match"] = match_condition.if_none_match
        if match_condition.if_version_match is not None:
            args["if_generation_match"] = int(match_condition.if_version_match)
        if match_condition.if_version_not_match is not None:
            args["if_generation_not_match"] = int(
                match_condition.if_version_not_match
            )
        return args

    def convert_copy_match_condition(
        self, match_condition: MatchCondition
    ) -> dict:
        args: dict = {}
        if match_condition.if_version_match is not None:
            args["if_source_generation_match"] = int(
                match_condition.if_version_match
            )
        elif match_condition.if_version_not_match is not None:
            args["if_source_generation_not_match"] = int(
                match_condition.if_version_not_match
            )
        return args

    def convert_put(
        self,
        id: str,
        value: bytes | None,
        file: str | None,
        stream: IO | None,
        metadata: dict | None,
        properties: dict | None,
        match_condition: MatchCondition,
        returning: bool | None,
        config: ObjectTransferConfig | None,
    ) -> tuple[
        dict,
        str,
        Blob,
    ]:
        blob = self.client.blob(blob_name=id)
        blob = self.convert_properties_metadata(
            blob=blob, properties=properties, metadata=metadata
        )
        if config is not None and config.multipart:
            func = "helper_put"
            args: dict = {
                "blob": blob,
                "value": value,
                "file": file,
                "stream": stream,
                "match_condition": match_condition,
                "returning": returning,
                "config": config,
            }
        else:
            args = {}
            if value is not None:
                func = "upload_from_file"
                args["file_obj"] = BytesIO(value)
            elif stream is not None:
                func = "upload_from_file"
                args["file_obj"] = stream
            elif file is not None:
                func = "upload_from_filename"
                args["filename"] = file
            args = args | self.convert_match_condition(match_condition)
        return args, func, blob

    def convert_get(
        self,
        id: str,
        version: str | None,
        file: str | None,
        stream: IO | None,
        match_condition: MatchCondition,
        start: int | None,
        end: int | None,
        config: ObjectTransferConfig | None,
    ) -> tuple[dict, str, Blob]:
        if version is None:
            blob = self.client.blob(blob_name=id)
        else:
            blob = self.client.blob(blob_name=id, generation=version)
        if config is not None and config.multipart:
            func = "helper_get"
            args: dict = {
                "blob": blob,
                "file": file,
                "stream": stream,
                "match_condition": match_condition,
                "config": config,
            }
        else:
            args = {}
            if start is not None:
                args["start"] = start
            if end is not None:
                args["end"] = end
            args = args | self.convert_match_condition(match_condition)
            if file is not None:
                args["filename"] = file
                func = "download_to_filename"
            elif stream is not None:
                args["file_obj"] = stream
                func = "download_to_file"
            else:
                func = "download_as_bytes"
            args = args | self.convert_match_condition(match_condition)
        return args, func, blob

    def convert_get_properties(
        self,
        id: str,
        version: str | None,
        match_condition: MatchCondition,
    ) -> dict:
        args: dict = {}
        args["blob_name"] = id
        if version is not None:
            args["generation"] = int(version)
        args = args | self.convert_match_condition(match_condition)
        return args

    def convert_get_versions(self, id: str) -> dict:
        return {"prefix": id, "versions": True}

    def convert_update(
        self,
        id: str,
        version: str | None,
        metadata: dict | None,
        properties: dict | None,
        match_condition: MatchCondition,
    ) -> tuple[dict, Blob, bool, bool]:
        update_storage_class = False
        update_others = False
        blob = self.client.blob(blob_name=id)
        if version is not None:
            blob.generation = int(version)
        blob = self.convert_properties_metadata(
            blob=blob, properties=properties, metadata=metadata
        )
        if blob.storage_class is not None:
            update_storage_class = True
            if (
                properties is not None and len(properties) > 1
            ) or metadata is not None:
                # updating more than storage class
                update_others = True
        else:
            update_others = True
        args = self.convert_match_condition(match_condition)
        return args, blob, update_storage_class, update_others

    def convert_delete(
        self,
        id: str,
        version: str | None,
        match_condition: MatchCondition,
    ) -> tuple[dict, str]:
        if version == "*":
            args: dict = {
                "id": id,
                "version": version,
                "match_condition": match_condition,
            }
            return args, "helper_delete"
        args = {"blob_name": id}
        if version is not None:
            args["generation"] = int(version)
        else:
            args["generation"] = 0
        args = args | self.convert_match_condition(match_condition)
        return args, "delete_blob"

    def convert_copy(
        self,
        id: str,
        source_id: str,
        source_version: str | None,
        source_collection: str | None,
        metadata: dict | None,
        properties: dict | None,
        match_condition: MatchCondition,
    ) -> tuple[dict, Bucket]:
        args: dict = {}
        source_bucket = self.storage_client.bucket(
            bucket_name=(
                source_collection
                if source_collection is not None
                else self.bucket
            )
        )
        blob = source_bucket.blob(blob_name=source_id)
        blob = self.convert_properties_metadata(blob, properties, metadata)
        args["blob"] = blob
        args["destination_bucket"] = self.client
        args["new_name"] = id
        if source_version is not None:
            args["source_generation"] = int(source_version)
        args = args | self.convert_copy_match_condition(match_condition)
        return args, source_bucket

    def convert_generate(
        self,
        id: str,
        version: str | None,
        method: str | None,
        expiry: int | None,
    ) -> tuple[dict, Blob]:
        args: dict = {}
        blob = self.client.blob(blob_name=id)
        if version is not None:
            args["generation"] = int(version)
        args["version"] = "v4"
        args["method"] = method if method is not None else "GET"
        args["expiration"] = timedelta(
            seconds=expiry if expiry is not None else 3600
        )
        return args, blob

    def convert_batch(
        self, op_parsers: list[StoreOperationParser]
    ) -> tuple[dict, str]:
        for op_parser in op_parsers:
            if not op_parser.op_equals(StoreOperation.DELETE):
                raise BadRequestError("Operation not supported in BATCH")
        return self.convert_batch_delete(op_parsers), "delete"

    def convert_batch_delete(
        self, op_parsers: list[StoreOperationParser]
    ) -> dict:
        objects = []
        for op_parser in op_parsers:
            id = op_parser.get_id_as_str()
            version = op_parser.get_version()
            objects.append({"id": id, "version": version})
        return {"objects": objects}


class GoogleCloudStorageCollection:
    client: Any
    helper: Any
    op_converter: OperationConverter
    result_converter: ResultConverter
    processor: ItemProcessor

    def __init__(
        self,
        storage_client: Any,
        client: Any,
        bucket: str,
    ):
        self.client = client
        self.processor = ItemProcessor()
        self.op_converter = OperationConverter(
            storage_client, self.client, self.processor, bucket
        )
        self.helper = ClientHelper(
            storage_client, client, bucket, self.op_converter
        )
        self.result_converter = ResultConverter(self.processor, bucket)



"""
Vector Store on Milvus.
"""

from __future__ import annotations

__all__ = ["Milvus"]

import json
from typing import Any

from pymilvus import DataType, MilvusClient

from verse.core import Context, NCall, Operation, Response
from verse.core.exceptions import BadRequestError, NotFoundError
from verse.internal.storage_core import (
    ItemProcessor,
    QueryProcessor,
    StoreFunctionName,
    StoreOperation,
    StoreOperationParser,
    StoreProvider,
    Validator,
)
from verse.ql import (
    And,
    Comparison,
    ComparisonOp,
    Expression,
    Field,
    Function,
    FunctionNamespace,
    Not,
    Or,
    OrderBy,
    Select,
)

from .._helper import get_collection_config, is_metadata_field, is_value_field
from .._models import (
    VectorCollectionConfig,
    VectorIndexMetric,
    VectorItem,
    VectorKey,
    VectorList,
    VectorProperties,
    VectorSearchArgs,
    VectorValue,
)

DEFAULT_LIMIT = 100


class Milvus(StoreProvider):
    uri: str
    token: str | None
    user: str | None
    password: str | None
    collection: str | None
    id_field: str
    vector_field: str
    metadata_field: str
    content_field: str | None
    sparse_vector_field: str | None
    nparams: dict[str, Any]

    _client: MilvusClient
    _processor: ItemProcessor

    _collection_cache: dict[str, MilvusCollection]

    def __init__(
        self,
        uri: str,
        token: str | None = None,
        user: str | None = None,
        password: str | None = None,
        collection: str | None = None,
        id_field: str = "id",
        vector_field: str = "vector",
        metadata_field: str = "metadata",
        content_field: str | None = "content",
        sparse_vector_field: str | None = None,
        nparams: dict[str, Any] = dict(),
        **kwargs,
    ):
        """Initialize.

        Args:
            uri:
                Milvus endpoint.
            token:
                Milvus API key.
            user:
                Milvus user name.
            password:
                Milvus password.
            collection:
                Milvus collection name mapped to vector store collection.
            id_field:
                Milvus field mapped as id attribute. Defaults to "id".
            vector_field:
                Milvus field mapped as vector value attribute.
                Defaults to "vector".
            metadata_field:
                Milvus field mapped as metadata attribute.
                Defaults to "metadata".
            content_field:
                Milvus field mapped as content value attribute.
                Defaults to "content".
            sparse_vector_field:
                Milvus field mapped as sparse vector value attribute.
                Defaults to None.
            nparams:
                Native parameters to milvus client.
        """
        self.uri = uri
        self.token = token
        self.user = user
        self.password = password
        self.collection = collection
        self.id_field = id_field
        self.vector_field = vector_field
        self.metadata_field = metadata_field
        self.content_field = content_field
        self.sparse_vector_field = sparse_vector_field
        self.nparams = nparams

        self._client = None
        self._processor = ItemProcessor()
        self._collection_cache = dict()

    def init(self, context: Context | None = None) -> None:
        if self._client is not None:
            return

        self._client = MilvusClient(
            uri=self.uri,
            token=self.token,
            user=self.user,
            password=self.password,
            **self.nparams,
        )

    def _get_collection_name(
        self, op_parser: StoreOperationParser
    ) -> str | None:
        collection_name = (
            op_parser.get_operation_parsers()[0].get_collection_name()
            if op_parser.op_equals(StoreOperation.BATCH)
            or op_parser.op_equals(StoreOperation.TRANSACT)
            else op_parser.get_collection_name()
        )
        collection = (
            collection_name if collection_name is not None else self.collection
        )
        return collection

    def _get_collection(
        self, op_parser: StoreOperationParser
    ) -> MilvusCollection | None:
        if op_parser.is_resource_op():
            return None
        collection_name = self._get_collection_name(op_parser)
        if collection_name is None:
            raise BadRequestError("Collection name must be specified")
        if collection_name in self._collection_cache:
            return self._collection_cache[collection_name]
        description = self._client.describe_collection(collection_name)
        dimension = self._parse_dimension(description)
        col = MilvusCollection(
            self._client,
            collection_name,
            dimension,
            self.id_field,
            self.vector_field,
            self.metadata_field,
            self.content_field,
            self.sparse_vector_field,
        )
        self._collection_cache[collection_name] = col
        return col

    def _validate(self, op_parser: StoreOperationParser):
        if op_parser.op_equals(StoreOperation.BATCH):
            Validator.validate_batch(
                op_parser.get_operation_parsers(),
                allowed_ops=[
                    StoreOperation.PUT,
                    StoreOperation.DELETE,
                    StoreOperation.GET,
                ],
                single_collection=True,
            )
        elif op_parser.op_equals(StoreOperation.TRANSACT):
            Validator.validate_transact(
                op_parser.get_operation_parsers(),
                allowed_ops=[
                    StoreOperation.PUT,
                    StoreOperation.DELETE,
                    StoreOperation.GET,
                ],
                single_collection=True,
            )

    def run(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
    ) -> Any:
        self.init(context=context)
        op_parser = self.get_op_parser(operation)
        self._validate(op_parser)
        collection = self._get_collection(op_parser)
        ncall, state = self._get_ncall(
            op_parser,
            collection,
            ResourceHelper(self._client),
        )
        if ncall is None:
            return super().run(operation)
        nresult = ncall.invoke()
        result = self._convert_nresult(nresult, state, op_parser, collection)
        return Response(result=result, native=dict(result=nresult, call=ncall))

    def _get_ncall(
        self,
        op_parser: StoreOperationParser,
        collection: MilvusCollection | None,
        resource_helper: Any,
    ) -> tuple[NCall | None, dict | None]:
        if collection is not None:
            converter = collection.converter
            client = collection.client
        call = None
        state = None
        nargs = op_parser.get_nargs()
        # CREATE COLLECTION
        if op_parser.op_equals(StoreOperation.CREATE_COLLECTION):
            args: dict[Any, Any] | None = {
                "collection": self._get_collection_name(op_parser),
                "config": get_collection_config(op_parser),
                "id_field": self.id_field,
                "vector_field": self.vector_field,
                "metadata_field": self.metadata_field,
                "content_field": self.content_field,
                "sparse_vector_field": self.sparse_vector_field,
                "nargs": nargs,
            }
            call = NCall(resource_helper.create_collection, args)
        # DROP COLLECTION
        elif op_parser.op_equals(StoreOperation.DROP_COLLECTION):
            args = {
                "collection": self._get_collection_name(op_parser),
                "nargs": nargs,
            }
            call = NCall(resource_helper.drop_collection, args)
        # LIST COLLECTION
        elif op_parser.op_equals(StoreOperation.LIST_COLLECTIONS):
            args = {"nargs": nargs}
            call = NCall(resource_helper.list_collections, args)
        # GET
        elif op_parser.op_equals(StoreOperation.GET):
            args = converter.convert_get(op_parser.get_id_as_str())
            call = NCall(client.get, args, nargs)
        # PUT
        elif op_parser.op_equals(StoreOperation.PUT):
            args = converter.convert_put(
                op_parser.get_id_as_str(),
                op_parser.get_vector_value(),
                op_parser.get_metadata(),
            )
            call = NCall(client.upsert, args, nargs)
        # DELETE
        elif op_parser.op_equals(StoreOperation.DELETE):
            args = converter.convert_delete(
                op_parser.get_id_as_str_or_none(), op_parser.get_where()
            )
            call = NCall(
                client.delete,
                args,
                nargs,
            )
        # QUERY
        elif op_parser.op_equals(StoreOperation.QUERY):
            args = converter.convert_query(
                search=op_parser.get_search_as_function(),
                where=op_parser.get_where(),
                select=op_parser.get_select(),
                order_by=op_parser.get_order_by(),
                limit=op_parser.get_limit(),
                offset=op_parser.get_offset(),
            )
            call = NCall(client.search, args, nargs)
        # COUNT
        elif op_parser.op_equals(StoreOperation.COUNT):
            args = converter.convert_count(
                where=op_parser.get_where(),
            )
            call = NCall(client.get_collection_stats, args, nargs)
        # BATCH or TRANSACT
        elif op_parser.op_equals(StoreOperation.BATCH) or op_parser.op_equals(
            StoreOperation.TRANSACT
        ):
            args, func = converter.convert_batch(
                op_parser.get_operation_parsers()
            )
            if func == "get":
                call = NCall(client.get, args, nargs)
            elif func == "put":
                call = NCall(client.upsert, args, nargs)
            elif func == "delete":
                call = NCall(client.delete, args, nargs)
            state = {"func": func}
        # CLOSE
        elif op_parser.op_equals(StoreOperation.CLOSE):
            args = {"nargs": nargs}
            call = NCall(resource_helper.close, args)
        return call, state

    def _convert_nresult(
        self,
        nresult: Any,
        state: dict | None,
        op_parser: StoreOperationParser,
        collection: MilvusCollection | None,
    ) -> Any:
        if collection is not None:
            processor = collection.processor
        result: Any = None
        # CREATE COLLECTION
        if op_parser.op_equals(StoreOperation.CREATE_COLLECTION):
            result = None
        # DROP COLLECTION
        elif op_parser.op_equals(StoreOperation.DROP_COLLECTION):
            result = None
        # LIST COLLECTION
        elif op_parser.op_equals(StoreOperation.LIST_COLLECTIONS):
            result = nresult
        # GET
        elif op_parser.op_equals(StoreOperation.GET):
            if len(nresult) == 0:
                raise NotFoundError
            result = self._convert_to_item(nresult[0], processor)
        # PUT
        elif op_parser.op_equals(StoreOperation.PUT):
            result = VectorItem(
                key=VectorKey(id=op_parser.get_id_as_str()),
            )
        # DELETE
        elif op_parser.op_equals(StoreOperation.DELETE):
            result = None
        # QUERY
        elif op_parser.op_equals(StoreOperation.QUERY):
            items = []
            for item in nresult[0]:
                items.append(self._convert_to_item(item, processor, True))
            items = QueryProcessor.order_items(
                items, op_parser.get_order_by(), processor.resolve_root_field
            )
            limit = op_parser.get_limit()
            items = QueryProcessor.limit_items(
                items,
                limit if limit is not None else DEFAULT_LIMIT,
                op_parser.get_offset(),
            )
            result = VectorList(items=items)
        # COUNT
        elif op_parser.op_equals(StoreOperation.COUNT):
            result = nresult["row_count"]
        # BATCH or TRANSACT
        elif op_parser.op_equals(StoreOperation.BATCH) or op_parser.op_equals(
            StoreOperation.TRANSACT
        ):
            result = []
            func = state["func"] if state is not None else None
            if func == "get":
                for item in nresult:
                    result.append(self._convert_to_item(item, processor))
            elif func == "put":
                for op_parser in op_parser.get_operation_parsers():
                    result.append(
                        VectorItem(key=VectorKey(id=op_parser.get_id_as_str()))
                    )
            elif func == "delete":
                for op_parser in op_parser.get_operation_parsers():
                    result.append(None)
        return result

    def _convert_to_item(
        self, nitem: Any, processor: ItemProcessor, entity: bool = False
    ) -> VectorItem:
        vector = None
        content = None
        sparse_vector = None
        metadata: dict | None = None
        distance: float | None = None
        if entity:
            distance = nitem["distance"]
            nitem = nitem["entity"]
        key = VectorKey(id=nitem[self.id_field])
        if self.vector_field in nitem:
            if entity:
                vector = nitem[self.vector_field]
            else:
                vector = [v.item() for v in nitem[self.vector_field]]
        if self.sparse_vector_field in nitem:
            sparse_vector = nitem[self.sparse_vector_field]
        if self.content_field in nitem:
            content = nitem[self.content_field]
        value = VectorValue(
            vector=vector,
            sparse_vector=sparse_vector,
            content=content,
        )

        if self.metadata_field in nitem:
            metadata = nitem[self.metadata_field]
            if isinstance(metadata, dict) and len(metadata) == 0:
                metadata = None
        properties = None
        if distance is not None:
            properties = VectorProperties(score=distance)
        return VectorItem(
            key=key, value=value, metadata=metadata, properties=properties
        )

    def _parse_dimension(self, description: Any) -> int:
        fields = description["fields"]
        for f in fields:
            if f["name"] == self.vector_field:
                return f["params"]["dim"]
        raise BadRequestError("Vector field not found")


class ResourceHelper:
    client: MilvusClient

    def __init__(self, client: MilvusClient):
        self.client = client

    def create_collection(
        self,
        collection: str,
        config: VectorCollectionConfig | None,
        id_field: str,
        vector_field: str,
        metadata_field: str,
        content_field: str | None,
        sparse_vector_field: str | None,
        nargs: Any,
    ):
        metric_mapping = {
            VectorIndexMetric.DOTPRODUCT: "IP",
            VectorIndexMetric.COSINE: "COSINE",
            VectorIndexMetric.EUCLIDEAN: "EUCLIDEAN",
        }
        vector_index = config.vector_index if config is not None else None
        if vector_index is not None:
            dimension = vector_index.dimension
            metric = metric_mapping[vector_index.metric]
        else:
            dimension = 4
            metric = "IP"

        schema = MilvusClient.create_schema(
            auto_id=False,
            enable_dynamic_field=False,
        )
        schema.add_field(
            field_name=id_field,
            datatype=DataType.VARCHAR,
            max_length=256,
            is_primary=True,
        )
        schema.add_field(
            field_name=vector_field,
            datatype=DataType.FLOAT_VECTOR,
            dim=dimension,
        )
        schema.add_field(field_name=metadata_field, datatype=DataType.JSON)
        if content_field is not None:
            schema.add_field(
                field_name=content_field,
                datatype=DataType.VARCHAR,
                max_length=65535,
            )
        if sparse_vector_field is not None:
            schema.add_field(
                field_name=sparse_vector_field,
                datatype=DataType.SPARSE_FLOAT_VECTOR,
            )
        index_params = self.client.prepare_index_params()
        index_params.add_index(field_name=id_field)
        index_params.add_index(
            field_name=vector_field,
            index_type="AUTOINDEX",
            metric_type=metric,
        )
        if sparse_vector_field is not None:
            index_params.add_index(field_name=sparse_vector_field)
        args = {
            "collection_name": collection,
            "schema": schema,
            "index_params": index_params,
        }
        return NCall(self.client.create_collection, args, nargs).invoke()

    def drop_collection(self, collection: str, nargs: Any):
        args = {
            "collection_name": collection,
        }
        return NCall(self.client.drop_collection, args, nargs).invoke()

    def list_collections(self, nargs) -> Any:
        return NCall(self.client.list_collections, None, nargs).invoke()

    def close(self, nargs: Any) -> Any:
        pass


class OperationConverter:
    processor: ItemProcessor
    collection: str
    dimension: int
    id_field: str
    vector_field: str
    metadata_field: str
    content_field: str | None
    sparse_vector_field: str | None

    def __init__(
        self,
        processor: ItemProcessor,
        collection: str,
        dimension: int,
        id_field: str,
        vector_field: str,
        metadata_field: str,
        content_field: str | None,
        sparse_vector_field: str | None,
    ) -> None:
        self.processor = processor
        self.collection = collection
        self.dimension = dimension
        self.id_field = id_field
        self.vector_field = vector_field
        self.metadata_field = metadata_field
        self.content_field = content_field
        self.sparse_vector_field = sparse_vector_field

    def convert_batch(
        self, op_parsers: list[StoreOperationParser]
    ) -> tuple[dict, str]:
        batch_get = False
        batch_put = False
        batch_delete = False
        for op_parser in op_parsers:
            if op_parser.op_equals(StoreOperation.GET):
                batch_get = True
            elif op_parser.op_equals(StoreOperation.PUT):
                batch_put = True
            elif op_parser.op_equals(StoreOperation.DELETE):
                batch_delete = True
        if (
            (batch_get and batch_put)
            or (batch_put and batch_delete)
            or (batch_delete and batch_get)
        ):
            raise BadRequestError(
                "Batch operation supports only one operation type"
            )
        if batch_get:
            return self.convert_batch_get(op_parsers), "get"
        elif batch_put:
            return self.convert_batch_put(op_parsers), "put"
        elif batch_delete:
            return self.convert_batch_delete(op_parsers), "delete"
        raise BadRequestError("Batch operation not supported")

    def convert_batch_get(
        self, op_parsers: list[StoreOperationParser]
    ) -> dict:
        ids = []
        for op_parser in op_parsers:
            ids.append(op_parser.get_id_as_str())
        return {"collection_name": self.collection, "ids": ids}

    def convert_batch_put(
        self, op_parsers: list[StoreOperationParser]
    ) -> dict:
        vectors = []
        for op_parser in op_parsers:
            vectors.append(
                self.convert_put_vector(
                    op_parser.get_id_as_str(),
                    op_parser.get_vector_value(),
                    op_parser.get_metadata(),
                )
            )
        return {
            "collection_name": self.collection,
            "data": vectors,
        }

    def convert_batch_delete(
        self, op_parsers: list[StoreOperationParser]
    ) -> dict:
        ids = []
        for op_parser in op_parsers:
            ids.append(op_parser.get_id_as_str())
        return {"collection_name": self.collection, "ids": ids}

    def convert_get(self, id: str) -> dict:
        return {"collection_name": self.collection, "ids": [id]}

    def convert_put(self, id: str, value: dict, metadata: dict | None) -> dict:
        return {
            "collection_name": self.collection,
            "data": [self.convert_put_vector(id, value, metadata)],
        }

    def convert_put_vector(
        self, id: str, value: dict, metadata: dict | None
    ) -> dict:
        vector_value = VectorValue(**value)
        args: dict = {
            self.id_field: id,
            self.vector_field: vector_value.vector,
        }
        if metadata is not None:
            args[self.metadata_field] = metadata
        else:
            args[self.metadata_field] = {}
        if self.content_field is not None and vector_value.content is not None:
            args[self.content_field] = vector_value.content
        if (
            self.sparse_vector_field is not None
            and vector_value.sparse_vector is not None
        ):
            args[self.sparse_vector_field] = vector_value.sparse_vector
        return args

    def convert_delete(self, id: str | None, where: Expression) -> dict:
        if id is None and where is None:
            return {"collection_name": self.collection}
        if id is not None and where is None:
            return {"collection_name": self.collection, "ids": [id]}
        if where is not None and id is None:
            return {
                "collection_name": self.collection,
                "filter": self.convert_expr(where),
            }
        raise BadRequestError(
            "Both id and where cannot be provided for delete"
        )

    def convert_query(
        self,
        search: Function | None,
        where: Expression,
        select: Select | None,
        order_by: OrderBy | None,
        limit: int | None,
        offset: int | None,
    ) -> dict:
        args: dict = {"collection_name": self.collection}
        search_args = self.convert_search(search)
        args = args | search_args
        select_args = self.convert_select(select)
        if select_args is not None:
            args = args | select_args
        if where is not None:
            args["filter"] = self.convert_expr(where)
        _offset = offset if offset is not None else 0
        _limit = limit if limit is not None else DEFAULT_LIMIT
        args["limit"] = _limit + _offset
        return args

    def convert_select(self, select: Select | None) -> dict:
        fields = [self.id_field]
        select_value = False
        if select is None or len(select.terms) == 0:
            select_value = True
            fields.append(self.metadata_field)
        else:
            for term in select.terms:
                if is_value_field(term.field):
                    select_value = True
                elif is_metadata_field(term.field):
                    fields.append(self.metadata_field)
        if select_value:
            fields.append(self.vector_field)
            if self.content_field is not None:
                fields.append(self.content_field)
            if self.sparse_vector_field is not None:
                fields.append(self.sparse_vector_field)
        return {"output_fields": fields}

    def convert_search(self, search: Function | None) -> dict:
        if search is None:
            return {"data": [[0.0] * self.dimension]}
        namespace = search.namespace
        name = search.name
        if namespace == FunctionNamespace.BUILTIN:
            if name == StoreFunctionName.VECTOR_SEARCH:
                args = VectorSearchArgs(**search.named_args)
                return {
                    "data": [args.vector],
                }

        raise BadRequestError("Search function not supported")

    def convert_expr(self, expr: Expression) -> str:
        if expr is None or isinstance(
            expr, (str, int, float, bool, dict, list)
        ):
            return json.dumps(expr)
        if isinstance(expr, Field):
            field = self.processor.resolve_metadata_field(expr.path)
            return f"{self.metadata_field}['{field}']"
        if isinstance(expr, Function):
            return self.convert_func(expr)
        if isinstance(expr, Comparison):
            comparion_map = {
                ComparisonOp.EQ: "==",
                ComparisonOp.NEQ: "!=",
                ComparisonOp.GT: ">",
                ComparisonOp.GTE: ">=",
                ComparisonOp.LT: "<",
                ComparisonOp.LTE: "<=",
                ComparisonOp.IN: "in",
                ComparisonOp.NIN: "not in",
                ComparisonOp.LIKE: "like",
            }
            op = expr.op
            if op in comparion_map:
                if isinstance(expr.lexpr, Field):
                    field = self.convert_expr(expr.lexpr)
                    value = self.convert_expr(expr.rexpr)
                elif isinstance(expr.rexpr, Field):
                    field = self.convert_expr(expr.rexpr)
                    value = self.convert_expr(expr.lexpr)
                    op = Comparison.reverse_op(op)
                m_op = comparion_map[op]
                return f"{field} {m_op} {value}"
            elif op == ComparisonOp.BETWEEN:
                field = self.convert_expr(expr.lexpr)
                if isinstance(expr.rexpr, list):
                    low = self.convert_expr(expr.rexpr[0])
                    high = self.convert_expr(expr.rexpr[1])
                    return f"{field} >= {low} and {field} <= {high}"
                raise BadRequestError("BETWEEN has bad value type")
            else:
                raise BadRequestError(f"Comparison op {expr.op} not supported")
        if isinstance(expr, And):
            lhs = self.convert_expr(expr.lexpr)
            rhs = self.convert_expr(expr.rexpr)
            return f"({lhs} and {rhs})"
        if isinstance(expr, Or):
            lhs = self.convert_expr(expr.lexpr)
            rhs = self.convert_expr(expr.rexpr)
            return f"({lhs} or {rhs})"
        if isinstance(expr, Not):
            hs = self.convert_expr(expr.expr)
            return f"not {hs}"
        return str(expr)

    def convert_func(self, expr: Function) -> str:
        namespace = expr.namespace
        name = expr.name
        args = expr.args
        if namespace == FunctionNamespace.BUILTIN:
            if name == StoreFunctionName.IS_DEFINED:
                field = self.convert_expr(args[0])
                return f"exists {field}"
            if name == StoreFunctionName.IS_NOT_DEFINED:
                field = self.convert_expr(args[0])
                return f"not exists {field}"
            if name == StoreFunctionName.ARRAY_CONTAINS:
                field = self.convert_expr(args[0])
                value = self.convert_expr(args[1])
                return f"json_contains({field}, {value})"
            if name == StoreFunctionName.ARRAY_CONTAINS_ANY:
                field = self.convert_expr(args[0])
                value = self.convert_expr(args[1])
                return f"json_contains_any({field}, {value})"
        raise BadRequestError(f"Function {name} not supported")

    def convert_count(self, where: Expression) -> dict | None:
        return {"collection_name": self.collection}


class MilvusCollection:
    client: Any
    converter: OperationConverter
    processor: ItemProcessor

    def __init__(
        self,
        client: Any,
        collection: str,
        dimension: int,
        id_field: str,
        vector_field: str,
        metadata_field: str,
        content_field: str | None,
        sparse_vector_field: str | None,
    ):
        self.client = client
        self.processor = ItemProcessor()
        self.converter = OperationConverter(
            processor=self.processor,
            collection=collection,
            dimension=dimension,
            id_field=id_field,
            vector_field=vector_field,
            metadata_field=metadata_field,
            content_field=content_field,
            sparse_vector_field=sparse_vector_field,
        )
